import atexit
import logging
from packaging.version import parse
import epics
from epics import ca, caget, caput

from ._dispatch import _CallbackThread, EventDispatcher, wrap_callback

_min_pyepics = "3.4.2"

if parse(epics.__version__) < parse(_min_pyepics):
    raise ImportError(
        "Version of pyepics too old. "
        f"Ophyd requires at least {_min_pyepics}"
        f"but {epics.__version__} is installed"
    )

try:
    ca.find_libca()
except ca.ChannelAccessException:
    raise ImportError("libca not found; pyepics is unavailable")
else:
    thread_class = ca.CAThread


module_logger = logging.getLogger(__name__)
name = "pyepics"
_dispatcher = None
get_pv = epics.get_pv


def get_dispatcher():
    "The event dispatcher for the pyepics control layer"
    return _dispatcher


class PyepicsCallbackThread(_CallbackThread):
    def attach_context(self):
        super().attach_context()
        ca.attach_context(self.context)

    def detach_context(self):
        super().detach_context()
        if ca.current_context() is not None:
            ca.detach_context()


class PyepicsShimPV(epics.PV):
    def __init__(
        self,
        pvname,
        callback=None,
        form="time",
        verbose=False,
        auto_monitor=None,
        count=None,
        connection_callback=None,
        connection_timeout=None,
        access_callback=None,
    ):
        connection_callback = wrap_callback(
            _dispatcher, "metadata", connection_callback
        )
        callback = wrap_callback(_dispatcher, "monitor", callback)
        access_callback = wrap_callback(_dispatcher, "metadata", access_callback)

        super().__init__(
            pvname,
            form=form,
            verbose=verbose,
            auto_monitor=auto_monitor,
            count=count,
            connection_timeout=connection_timeout,
            connection_callback=connection_callback,
            callback=callback,
            access_callback=access_callback,
        )

        self._cache_key = (pvname, form, self.context)
        self._reference_count = 0

    def add_callback(
        self, callback=None, index=None, run_now=False, with_ctrlvars=True, **kw
    ):
        if not self.auto_monitor:
            self.auto_monitor = ca.DEFAULT_SUBSCRIPTION_MASK
        callback = wrap_callback(_dispatcher, "monitor", callback)
        return super().add_callback(
            callback=callback,
            index=index,
            run_now=run_now,
            with_ctrlvars=with_ctrlvars,
            **kw,
        )

    def put(
        self,
        value,
        wait=False,
        timeout=None,
        use_complete=False,
        callback=None,
        callback_data=None,
    ):
        callback = wrap_callback(_dispatcher, "get_put", callback)
        # pyepics does not accept an indefinite timeout
        if timeout is None:
            timeout = 315569520  # ten years
        return super().put(
            value,
            wait=wait,
            timeout=timeout,
            use_complete=use_complete,
            callback=callback,
            callback_data=callback_data,
        )

    def _getarg(self, arg):
        "wrapper for property retrieval"
        # NOTE: replaces epics.PV._getarg: does not call get() when value unset
        if self._args[arg] is None:
            if arg in (
                "status",
                "severity",
                "timestamp",
                "posixseconds",
                "nanoseconds",
            ):
                self.get_timevars(timeout=1, warn=False)
            else:
                self.get_ctrlvars(timeout=1, warn=False)
        return self._args.get(arg, None)

    def get_all_metadata_blocking(self, timeout):
        if self._args["status"] is None:
            self.get_timevars(timeout=timeout)
        self.get_ctrlvars(timeout=timeout)
        md = self._args.copy()
        md.pop("value", None)
        return md

    def get_all_metadata_callback(self, callback, *, timeout):
        def get_metadata_thread(pvname):
            md = self.get_all_metadata_blocking(timeout=timeout)
            callback(pvname, md)

        _dispatcher.schedule_utility_task(get_metadata_thread, pvname=self.pvname)

    def clear_callbacks(self):
        super().clear_callbacks()
        self.access_callbacks.clear()
        self.connection_callbacks.clear()


def release_pvs(*pvs):
    for pv in pvs:
        pv._reference_count -= 1
        if pv._reference_count == 0:
            pv.clear_callbacks()
            pv.clear_auto_monitor()
            if pv.chid is not None:
                # Clear the channel on the CA-level
                epics.ca.clear_channel(pv.chid)

            pv.chid = None
            pv.context = None

            # Ensure we don't get this same PV back again
            epics.pv._PVcache_.pop(pv._cache_key, None)


def setup(logger):
    """Setup ophyd for use

    Must be called once per session using ophyd
    """
    # It's important to use the same context in the callback _dispatcher
    # as the main thread, otherwise not-so-savvy users will be very
    # confused
    global _dispatcher

    if _dispatcher is not None:
        logger.debug("ophyd already setup")
        return

    epics.pv.default_pv_class = PyepicsShimPV

    def _cleanup():
        """Clean up the ophyd session"""
        global _dispatcher
        if _dispatcher is None:
            return
        epics.pv.default_pv_class = epics.PV

        if _dispatcher.is_alive():
            _dispatcher.stop()

        _dispatcher = None

    logger.debug("Installing event dispatcher")
    _dispatcher = EventDispatcher(
        thread_class=PyepicsCallbackThread, context=ca.current_context(), logger=logger
    )
    atexit.register(_cleanup)
    return _dispatcher


__all__ = (
    "setup",
    "caput",
    "caget",
    "get_pv",
    "thread_class",
    "name",
    "release_pvs",
    "get_dispatcher",
)



# vi: ts=4 sw=4
import os
import time
import threading
import warnings
import weakref

import numpy as np

from .utils import ReadOnlyError, LimitError, DestroyedError, doc_annotation_forwarder
from .utils.epics_pvs import (
    waveform_to_string,
    raise_if_disconnected,
    data_type,
    data_shape,
    AlarmStatus,
    AlarmSeverity,
    validate_pv_name,
    _set_and_wait,
)
from .ophydobj import OphydObject, Kind
from .status import Status
from . import get_cl


# Catch semi-frequent issue with scripts accidentally run from inside module
if __name__ != "ophyd.signal":
    raise RuntimeError(
        "A script tried to import ophyd.signal instead of the signal built-in "
        "module. This usually happens when a script is run from inside the "
        "ophyd directory and can cause extremely confusing bugs. Please "
        "run your script elsewhere for better results."
    )


# Sentinels used for default values; see set_defaults() below for details.
DEFAULT_AUTO_MONITOR = object()
DEFAULT_CONNECTION_TIMEOUT = object()
DEFAULT_TIMEOUT = object()
DEFAULT_WRITE_TIMEOUT = object()

# Sentinel to identify if we have never turned the crank on updating a PV
DEFAULT_EPICSSIGNAL_VALUE = object()


class ReadTimeoutError(TimeoutError):
    ...


class ConnectionTimeoutError(TimeoutError):
    ...


class Signal(OphydObject):
    r"""A signal, which can have a read-write or read-only value.

    Parameters
    ----------
    name : string, keyword only
    value : any, optional
        The initial value
    kind : a member the Kind IntEnum (or equivalent integer), optional
        Default is Kind.normal. See Kind for options.
    parent : Device, optional
        The parent Device holding this signal
    timestamp : float, optional
        The timestamp associated with the initial value. Defaults to the
        current local time.
    tolerance : any, optional
        The absolute tolerance associated with the value
    rtolerance : any, optional
        The relative tolerance associated with the value, used in
        set as follows

        .. math::

          |setpoint - readback| \leq (tolerance + rtolerance * |readback|)

    cl : namespace, optional
        Control Layer.  Must provide 'get_pv' and 'thread_class'
    attr_name : str, optional
        The parent Device attribute name that corresponds to this Signal

    Attributes
    ----------
    rtolerance : any, optional
        The relative tolerance associated with the value
    """
    SUB_VALUE = "value"
    SUB_META = "meta"
    _default_sub = SUB_VALUE
    _metadata_keys = None
    _core_metadata_keys = ("connected", "read_access", "write_access", "timestamp")

    def __init__(
        self,
        *,
        name,
        value=0.0,
        timestamp=None,
        parent=None,
        labels=None,
        kind=Kind.hinted,
        tolerance=None,
        rtolerance=None,
        metadata=None,
        cl=None,
        attr_name="",
    ):

        super().__init__(
            name=name, parent=parent, kind=kind, labels=labels, attr_name=attr_name
        )

        if cl is None:
            cl = get_cl()
        self.cl = cl
        self._dispatcher = cl.get_dispatcher()
        self._metadata_thread_ctx = self._dispatcher.get_thread_context("monitor")
        self._readback = value

        if timestamp is None:
            timestamp = time.time()

        self._destroyed = False

        self._set_thread = None
        self._tolerance = tolerance
        # self.tolerance is a property
        self.rtolerance = rtolerance

        # Signal defaults to being connected, with full read/write access.
        # Subclasses are expected to clear these on init, if applicable.
        self._metadata = dict(
            connected=True,
            read_access=True,
            write_access=True,
            timestamp=timestamp,
            status=None,
            severity=None,
            precision=None,
        )

        if metadata is not None:
            self._metadata.update(**metadata)

        if self._metadata_keys is None:
            self._metadata_keys = tuple(self._metadata.keys())
        else:
            unset_metadata = {
                key: None for key in self._metadata_keys if key not in self._metadata
            }

            self._metadata.update(**unset_metadata)

    def trigger(self):
        """Call that is used by bluesky prior to read()"""
        # NOTE: this is a no-op that exists here for bluesky purposes
        #       it may need to be moved in the future
        d = Status(self)
        d._finished()
        return d

    def wait_for_connection(self, timeout=0.0):
        """Wait for the underlying signals to initialize or connect"""
        pass

    @property
    def metadata_keys(self):
        "Metadata keys that will be passed along on value subscriptions"
        return tuple(self._metadata_keys)

    @property
    def timestamp(self):
        """Timestamp of the readback value"""
        return self._metadata["timestamp"]

    @property
    def tolerance(self):
        """The absolute tolerance associated with the value."""
        return self._tolerance

    @tolerance.setter
    def tolerance(self, tolerance):
        self._tolerance = tolerance

    def _repr_info(self):
        "Yields pairs of (key, value) to generate the Signal repr"
        yield from super()._repr_info()
        try:
            value = self._readback
        except Exception:
            value = None

        if value is not DEFAULT_EPICSSIGNAL_VALUE:
            yield ("value", value)

        yield ("timestamp", self._metadata["timestamp"])

        if self.tolerance is not None:
            yield ("tolerance", self.tolerance)

        if self.rtolerance is not None:
            yield ("rtolerance", self.rtolerance)

        # yield ('metadata', self._metadata)

    def get(self, **kwargs):
        """The readback value"""
        return self._readback

    def put(
        self,
        value,
        *,
        timestamp=None,
        force=False,
        metadata=None,
        timeout=DEFAULT_WRITE_TIMEOUT,
        **kwargs,
    ):
        """
        Low-level method for writing to a Signal.

        The value is optionally checked first, depending on the value of force.
        In addition, VALUE subscriptions are run.

        Extra kwargs are ignored (for API compatibility with EpicsSignal kwargs
        pass through).

        Parameters
        ----------
        value : any
            Value to set
        timestamp : float, optional
            The timestamp associated with the value, defaults to time.time()
        metadata : dict, optional
            Further associated metadata with the value (such as alarm status,
            severity, etc.)
        force : bool, optional
            Check the value prior to setting it, defaults to False

        """
        self.control_layer_log.debug(
            "put(value=%s, timestamp=%s, force=%s, metadata=%s)",
            value,
            timestamp,
            force,
            metadata,
        )

        if kwargs:
            warnings.warn(
                "Signal.put no longer takes keyword arguments; "
                "These are ignored and will be deprecated. "
                f"Received kwargs={kwargs}"
            )

        if not force:
            if not self.write_access:
                raise ReadOnlyError("Signal does not allow write access")

            self.check_value(value)

        old_value = self._readback
        self._readback = value

        if metadata is None:
            metadata = {}

        if timestamp is None:
            timestamp = metadata.get("timestamp", time.time())

        metadata = metadata.copy()
        metadata["timestamp"] = timestamp
        self._metadata.update(**metadata)

        md_for_callback = {
            key: metadata[key] for key in self._metadata_keys if key in metadata
        }

        if "timestamp" not in self._metadata_keys:
            md_for_callback["timestamp"] = timestamp

        self._run_subs(
            sub_type=self.SUB_VALUE, old_value=old_value, value=value, **md_for_callback
        )

    def _set_and_wait(self, value, timeout, **kwargs):
        """
        Overridable hook for subclasses to override :meth:`.set` functionality.

        This will be called in a separate thread (`_set_thread`), but will not
        be called in parallel.

        Parameters
        ----------
        value : any
            The value
        timeout : float, optional
            Maximum time to wait for value to be successfully set, or None
        """
        return _set_and_wait(
            self,
            value,
            timeout=timeout,
            atol=self.tolerance,
            rtol=self.rtolerance,
            **kwargs,
        )

    def set(self, value, *, timeout=None, settle_time=None, **kwargs):
        """
        Set the value of the Signal and return a Status object.

        Returns
        -------
        st : Status
            This status object will be finished upon return in the
            case of basic soft Signals
        """
        self.log.debug(
            "set(value=%s, timeout=%s, settle_time=%s, kwargs=%s)",
            value,
            timeout,
            settle_time,
            kwargs,
        )

        def set_thread():
            try:
                self._set_and_wait(value, timeout, **kwargs)
            except TimeoutError:
                success = False
                self.log.warning(
                    "%s: _set_and_wait(value=%s, timeout=%s, atol=%s, rtol=%s, kwargs=%s)",
                    self.name,
                    value,
                    timeout,
                    self.tolerance,
                    self.rtolerance,
                    kwargs,
                )
            except Exception:
                success = False
                self.log.exception(
                    "%s: _set_and_wait(value=%s, timeout=%s, atol=%s, rtol=%s, kwargs=%s)",
                    self.name,
                    value,
                    timeout,
                    self.tolerance,
                    self.rtolerance,
                    kwargs,
                )
            else:
                success = True
                self.log.debug(
                    "%s: _set_and_wait(value=%s, timeout=%s, atol=%s, rtol=%s, kwargs=%s) succeeded => %s",
                    self.name,
                    value,
                    timeout,
                    self.tolerance,
                    self.rtolerance,
                    kwargs,
                    self._readback,
                )

                if settle_time is not None:
                    self.log.debug("settling for %d seconds", settle_time)
                    time.sleep(settle_time)
            finally:
                # keep a local reference to avoid any GC shenanigans
                th = self._set_thread
                # these two must be in this order to avoid a race condition
                self._set_thread = None
                st._finished(success=success)
                del th

        if self._set_thread is not None:
            raise RuntimeError(
                "Another set() call is still in progress " f"for {self.name}"
            )

        st = Status(self)
        self._status = st
        self._set_thread = self.cl.thread_class(target=set_thread)
        self._set_thread.daemon = True
        self._set_thread.start()
        return self._status

    @property
    def value(self):
        """The signal's value"""
        fix_msg = (
            "We are falling back to calling `.get` and interrogating "
            "the underlying control system, however this may cause several "
            "other problems:\n"
            "   1. This property access may take an arbitrarily long time\n"
            "   2. This property access, which you expect to be read only "
            "may change other state in the Signal.\n"
            "Your options to fix this are:\n"
            "  - do not use obj.value.\n"
            "    - If you are using this is in a plan you "
            "like want to be using bps.read, bps.rd, bpp.reset_positions_decorator, "
            "bpp.reset_positions_wrapper, bpp.relative_set_decorator, or "
            "bpp.relative_set_wrapper\n"
            "    - if you are doing this in an ophyd method use `self.get`\n"
            "  - set up the Signal to monitor\n\n"
            "This behavior will likely change in the future."
        )

        if self._readback is DEFAULT_EPICSSIGNAL_VALUE:
            # If we are here, then we have never turned the crank on this Signal.  The current
            # behavior is to fallback to poking the control system to get the value, however this
            # is problematic and we may want to change in the future so warn verbosely
            if not os.getenv("OPHYD_SILENCE_VALUE_WARNING") == "1":
                warnings.warn(
                    f"You have called obj.value on {self} "
                    f"({self.name}.{self.dotted_name}) "
                    "which has not gotten value from the control system yet.\n"
                    + fix_msg,
                    stacklevel=2,
                )
            return self.get()
        else:
            # if we are in here then we have put/get at least once and/or are monitored
            has_monitors = hasattr(self, "_monitors") and all(
                v is not None for v in self._monitors.values()
            )
            if not has_monitors:
                # If we are not monitored, then warn that this may change in the future.
                if not os.getenv("OPHYD_SILENCE_VALUE_WARNING") == "1":
                    warnings.warn(
                        f"You have called obj.value on {self} "
                        f"({self.name}.{self.dotted_name}) "
                        "which is a non-monitored signal.\n" + fix_msg,
                        stacklevel=2,
                    )
                return self.get()

            # else return our cached value and assume something else is keeping us up-to-date
            # so we can trust the latest news
            return self._readback

    @value.setter
    def value(self, value):
        self.put(value)

    @raise_if_disconnected
    def read(self):
        """Put the status of the signal into a simple dictionary format
        for data acquisition

        Returns
        -------
            dict
        """
        value = self.get()
        return {self.name: {"value": value, "timestamp": self.timestamp}}

    def describe(self):
        """Provide schema and meta-data for :meth:`~BlueskyInterface.read`

        This keys in the `OrderedDict` this method returns must match the
        keys in the `OrderedDict` return by :meth:`~BlueskyInterface.read`.

        This provides schema related information, (ex shape, dtype), the
        source (ex PV name), and if available, units, limits, precision etc.

        Returns
        -------
        data_keys : OrderedDict
            The keys must be strings and the values must be dict-like
            with the ``event_model.event_descriptor.data_key`` schema.
        """
        if self._readback is DEFAULT_EPICSSIGNAL_VALUE:
            val = self.get()
        else:
            val = self._readback
        try:
            return {
                self.name: {
                    "source": "SIM:{}".format(self.name),
                    "dtype": data_type(val),
                    "shape": data_shape(val),
                }
            }
        except ValueError as ve:
            # data_type(val) raises ValueError if type(val) is not bluesky-friendly
            # help the humans by reporting self.name in the exception chain
            raise ValueError(
                f"failed to describe '{self.name}' with value '{val}'"
            ) from ve

    def read_configuration(self):
        "Dictionary mapping names to value dicts with keys: value, timestamp"
        return self.read()

    def describe_configuration(self):
        """Provide schema & meta-data for `~BlueskyInterface.read_configuration`

        This keys in the `OrderedDict` this method returns must match the keys
        in the `OrderedDict` return by :meth:`~BlueskyInterface.read`.

        This provides schema related information, (ex shape, dtype), the source
        (ex PV name), and if available, units, limits, precision etc.

        Returns
        -------
        data_keys : OrderedDict
            The keys must be strings and the values must be dict-like
            with the ``event_model.event_descriptor.data_key`` schema.
        """
        return self.describe()

    @property
    def limits(self):
        """The control limits (low, high), such that low <= value <= high"""
        # NOTE: subclasses are expected to override this property
        # Always override, never extend this
        return (0, 0)

    @property
    def low_limit(self):
        "The low, inclusive control limit for the Signal"
        return self.limits[0]

    @property
    def high_limit(self):
        "The high, inclusive control limit for the Signal"
        return self.limits[1]

    @property
    def hints(self):
        "Field hints for plotting"
        if (~Kind.normal & Kind.hinted) & self.kind:
            return {"fields": [self.name]}
        else:
            return {"fields": []}

    @property
    def connected(self):
        "Is the signal connected to its associated hardware, and ready to use?"
        return self._metadata["connected"] and not self._destroyed

    @property
    def read_access(self):
        "Can the signal be read?"
        return self._metadata["read_access"]

    @property
    def write_access(self):
        "Can the signal be written to?"
        return self._metadata["write_access"]

    @property
    def metadata(self):
        "A copy of the metadata dictionary associated with the signal"
        return self._metadata.copy()

    def destroy(self):
        """Disconnect the Signal from the underlying control layer; destroy it

        Clears all subscriptions on this Signal.  Once destroyed, the signal
        may no longer be used.
        """
        self._destroyed = True
        super().destroy()

    def _run_metadata_callbacks(self):
        "Run SUB_META in the appropriate dispatcher thread"
        self._metadata_thread_ctx.run(
            self._run_subs, sub_type=self.SUB_META, **self._metadata
        )


class SignalRO(Signal):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._metadata.update(
            connected=True,
            write_access=False,
        )

    def put(self, value, *, timestamp=None, force=False):
        raise ReadOnlyError("The signal {} is readonly.".format(self.name))

    def set(self, value, *, timestamp=None, force=False):
        raise ReadOnlyError("The signal {} is readonly.".format(self.name))


class DerivedSignal(Signal):
    def __init__(
        self, derived_from, *, write_access=None, name=None, parent=None, **kwargs
    ):
        """A signal which is derived from another one

        Calculations of the DerivedSignal value can be done in subclasses of
        DerivedSignal, overriding the `forward` and `inverse` methods.

        Metadata keys and write access are inherited from the main signal,
        referred to as `derived_from`.

        The description of this Signal, from `describe` will include an
        additional key indicating the signal name from where it was derived.

        Parameters
        ----------
        derived_from : Union[Signal, str]
            The signal from which this one is derived.  This may be a string
            attribute name that indicates a sibling to use.  When used in a
            Device, this is then simply the attribute name of another
            Component.
        name : str, optional
            The signal name
        parent : Device, optional
            The parent device
        """
        if isinstance(derived_from, str):
            derived_from = getattr(parent, derived_from)

        # Metadata keys from the class itself take precedence
        self._metadata_keys = getattr(self, "_metadata_keys", None)

        # However, if not specified, the keys from the original signal are used
        if self._metadata_keys is None:
            self._metadata_keys = getattr(derived_from, "metadata_keys", None)
            # And failing that, they are the defaults from all signals

        super().__init__(
            name=name, parent=parent, metadata=derived_from.metadata, **kwargs
        )

        self._derived_from = derived_from

        self._allow_writes = write_access is not False
        self._metadata["write_access"] = (
            derived_from.write_access and self._allow_writes
        )

        if self.connected:
            # set up the initial timestamp reporting, if connected
            self._metadata["timestamp"] = derived_from.timestamp

        derived_from.subscribe(
            self._derived_value_callback, event_type=self.SUB_VALUE, run=self.connected
        )
        derived_from.subscribe(
            self._derived_metadata_callback,
            event_type=self.SUB_META,
            run=self.connected,
        )

    @property
    def derived_from(self):
        """Signal that this one is derived from"""
        return self._derived_from

    def describe(self):
        """Description based on the original signal description"""
        desc = super().describe()[self.name]  # Description of this signal
        desc["derived_from"] = self._derived_from.name
        # Description of the derived signal
        derived_desc = self._derived_from.describe()[self._derived_from.name]
        derived_desc.update(desc)
        return {self.name: derived_desc}

    def _update_metadata_from_callback(self, **kwargs):
        updated_md = {key: kwargs[key] for key in self.metadata_keys if key in kwargs}

        if "write_access" in updated_md:
            updated_md["write_access"] = (
                updated_md["write_access"] and self._allow_writes
            )
        self._metadata.update(**updated_md)
        return updated_md

    def _derived_metadata_callback(
        self, *, connected, read_access, write_access, timestamp, **kwargs
    ):
        "Main signal metadata updated - update the DerivedSignal"
        self._update_metadata_from_callback(
            connected=connected,
            read_access=read_access,
            write_access=write_access,
            timestamp=timestamp,
            **kwargs,
        )

        self._run_metadata_callbacks()

    def _derived_value_callback(self, value, **kwargs):
        "Main signal value updated - update the DerivedSignal"
        # if some how we get cycled with the default value sentinel, just bail
        if value is DEFAULT_EPICSSIGNAL_VALUE:
            return
        value = self.inverse(value)
        self._readback = value
        updated_md = self._update_metadata_from_callback(**kwargs)
        self._run_subs(sub_type=self.SUB_VALUE, value=value, **updated_md)

    def get(self, **kwargs):
        "Get the value from the original signal, with `inverse` applied to it"
        value = self._derived_from.get(**kwargs)
        self._readback = self.inverse(value)
        self._metadata["timestamp"] = self._derived_from.timestamp
        return self._readback

    def inverse(self, value):
        """Compute original signal value -> derived signal value"""
        return value

    def put(self, value, **kwargs):
        """Put the value to the original signal"""
        if not self.write_access:
            raise ReadOnlyError("DerivedSignal is marked as read-only")
        value = self.forward(value)
        res = self._derived_from.put(value, **kwargs)
        self._metadata["timestamp"] = self._derived_from.timestamp
        return res

    def forward(self, value):
        """Compute derived signal value -> original signal value"""
        return value

    def wait_for_connection(self, timeout=0.0):
        """Wait for the original signal to connect"""
        return self._derived_from.wait_for_connection(timeout=timeout)

    @property
    def connected(self):
        """Mirrors the connection state of the original signal"""
        return self._derived_from.connected

    @property
    def limits(self):
        """Limits from the original signal (low, high), such that low <= value <= high"""
        return tuple(sorted(self.inverse(v) for v in self._derived_from.limits))

    def _repr_info(self):
        "Yields pairs of (key, value) to generate the Signal repr"
        yield from super()._repr_info()
        if self.parent is not None:
            yield ("derived_from", self._derived_from.dotted_name)
        else:
            yield ("derived_from", self._derived_from)


class InternalSignalMixin:
    """
    Mix-in class for adding the `InternalSignal` behavior to any signal class.

    A signal class with this mixin will reject all sets and puts unless
    internal=True is passed as an argument.

    The intended use for this is to signify that a signal is for internal use
    by the class only. That is, it would be a mistake to try to cause puts to
    this signal by code external to the Device class.

    Some more concrete use-cases would be things like soft "status" type
    signals that should be read-only except that the class needs to edit it,
    or EPICS signals that should be written to by the class but are likely to
    cause issues for external writes due to behavior complexity.
    """

    def put(self, *args, internal: bool = False, **kwargs):
        """
        Write protection for an internal signal.

        This method is not intended to be used from outside of the device
        that defined this signal. All writes must be done with internal=True.
        """
        if not internal:
            raise InternalSignalError()
        return super().put(*args, **kwargs)

    def set(self, *args, internal: bool = False, **kwargs):
        """
        Write protection for an internal signal.

        This method is not intended to be used from outside of the device
        that defined this signal. All writes must be done with internal=True.
        """
        if not internal:
            raise InternalSignalError()
        return super().set(*args, internal=internal, **kwargs)


class InternalSignal(InternalSignalMixin, Signal):
    """
    A soft Signal that stores data but should only be updated by the Device.

    Unlike SignalRO, which will unilaterally block all writes, this will
    allow writes with internal=True.

    The intended use for this is to signify that a signal is for internal use
    by the class only. That is, it would be a mistake to try to cause puts to
    this signal by code external to the Device class.

    Some more concrete use-cases would be things like soft "status" type
    signals that should be read-only except that the class needs to edit it,
    or calculated "done" signals for positioner classes.
    """


class InternalSignalError(ReadOnlyError):
    """
    A read-only error sourced from trying to write to an internal signal.
    """

    def __init__(self, message=None):
        if message is None:
            message = (
                "This signal is for internal use only. "
                "You should not be writing to it from outside "
                "the parent class. If you do need to write to "
                "this signal, you can use signal.put(value, internal=True)."
            )
        super().__init__(message)


class EpicsSignalBase(Signal):
    """A read-only EpicsSignal -- that is, one with no `write_pv`

    Keyword arguments are passed on to the base class (Signal) initializer

    Parameters
    ----------
    read_pv : str
        The PV to read from
    string : bool, optional
        Attempt to cast the EPICS PV value to a string by default
    auto_monitor : bool, optional
        Use automonitor with epics.PV
    name : str, optional
        Name of signal.  If not given defaults to read_pv
    metadata : dict
        Merged with metadata received from EPICS
    all_pvs : set
        Set of PVs to watch for connection and access rights callbacks.
        Defaults to ``{read_pvs}``.
    timeout : float or None, optional
        The timeout for serving a read request on a connected channel. This is
        only applied if the PV is connected within connection_timeout (below).

        The default value DEFAULT_TIMEOUT means, "Fall back to class-wide
        default." See EpicsSignalBase.set_defaults to configure class
        defaults.

        Explicitly passing None means, "Wait forever."
    write_timeout : float or None, optional
        The timeout for a reply when put completion is used. This is
        only applied if the PV is connected within connection_timeout (below).

        This is very different than the connection and read timeouts
        above. It relates to how long an action takes to complete, such motor
        motion or data acquisition. Any default value we choose here is likely
        to cause problems---either by being too short and giving up too early
        on a lengthy action or being too long and delaying the report of a
        failure. A finite value can be injected here or, perhaps more usefully,
        via `set` at the Device level, where a context-appropriate value can be
        chosen.
    connection_timeout : float or None, optional
        Timeout for connection. This includes the time to search and establish
        a channel.

        The default value DEFAULT_CONNECTION_TIMEOUT means, "Fall back to
        class-wide default." See EpicsSignalBase.set_defaults to
        configure class defaults.

        Explicitly passing None means, "Wait forever."
    """

    # This is set to True when the first instance is made. It is used to ensure
    # that certain class-global settings can only be made before any
    # instantiation.
    __any_instantiated = False

    # See set_defaults() for more on these.
    __default_connection_timeout = 1.0
    __default_timeout = 2.0  # *read* timeout
    __default_write_timeout = None  # Wait forever.
    __default_auto_monitor = False

    _read_pv_metadata_key_map = dict(
        status=("status", AlarmStatus),
        severity=("severity", AlarmSeverity),
        precision=("precision", None),
        lower_ctrl_limit=("lower_ctrl_limit", None),
        upper_ctrl_limit=("upper_ctrl_limit", None),
        timestamp=("timestamp", None),
        units=("units", None),
        enum_strs=("enum_strs", tuple),
        # ignored: read_access, write_access, connected, etc.
    )

    _metadata_keys = Signal._core_metadata_keys + (
        "status",
        "severity",
        "precision",
        "lower_ctrl_limit",
        "upper_ctrl_limit",
        "units",
        "enum_strs",
    )

    def __init__(
        self,
        read_pv,
        *,
        string=False,
        auto_monitor=DEFAULT_AUTO_MONITOR,
        name=None,
        metadata=None,
        all_pvs=None,
        timeout=DEFAULT_TIMEOUT,
        write_timeout=DEFAULT_WRITE_TIMEOUT,
        connection_timeout=DEFAULT_CONNECTION_TIMEOUT,
        **kwargs,
    ):
        self._metadata_lock = threading.RLock()
        self._read_pv = None
        self._read_pvname = read_pv
        self._string = bool(string)

        self._signal_is_ready = threading.Event()
        self._first_connection = True

        if auto_monitor is DEFAULT_AUTO_MONITOR:
            auto_monitor = self.__default_auto_monitor
        self._auto_monitor = auto_monitor
        if connection_timeout is DEFAULT_CONNECTION_TIMEOUT:
            connection_timeout = self.__default_connection_timeout
        self._connection_timeout = connection_timeout
        if timeout is DEFAULT_TIMEOUT:
            timeout = self.__default_timeout
        self._timeout = timeout
        if write_timeout is DEFAULT_WRITE_TIMEOUT:
            write_timeout = self.__default_write_timeout
        self._write_timeout = write_timeout

        if name is None:
            name = read_pv

        if metadata is None:
            metadata = {}

        metadata.update(
            connected=False,
        )

        kwargs.setdefault("value", DEFAULT_EPICSSIGNAL_VALUE)
        super().__init__(name=name, metadata=metadata, **kwargs)

        validate_pv_name(read_pv)

        # Keep track of all associated PV's connectivity and access rights
        # callbacks. These map `pvname` to bool:
        if all_pvs is None:
            all_pvs = {read_pv}
        self._connection_states = {pv: False for pv in all_pvs}
        self._access_rights_valid = {pv: False for pv in all_pvs}
        self._received_first_metadata = {pv: False for pv in all_pvs}
        self._monitors = {pv: None for pv in all_pvs}

        self._metadata_key_map = {read_pv: self._read_pv_metadata_key_map}

        for pv in all_pvs:
            if pv not in self._metadata_key_map:
                self._metadata_key_map[pv] = {}

        self._read_pv = self.cl.get_pv(
            read_pv,
            auto_monitor=self._auto_monitor,
            connection_callback=self._pv_connected,
            access_callback=self._pv_access_callback,
        )
        self._read_pv._reference_count += 1

        if not self.__any_instantiated:
            self.log.debug(
                "This is the first instance of EpicsSignalBase. "
                "name={self.name}, id={id(self)}"
            )
            EpicsSignalBase._mark_as_instantiated()

        def finalize(read_pv, cl):
            cl.release_pvs(read_pv)

        self._read_pv_finalizer = weakref.finalize(
            self, finalize, self._read_pv, self.cl
        )

    def destroy(self):
        super().destroy()
        self._read_pv_finalizer()

    @classmethod
    def _mark_as_instantiated(cls):
        "Update state indicated that this class has been instantiated."
        cls.__any_instantiated = True

    @classmethod
    def set_defaults(
        cls,
        *,
        timeout=__default_timeout,
        connection_timeout=__default_connection_timeout,
        write_timeout=__default_write_timeout,
        auto_monitor=__default_auto_monitor,
    ):
        """
        Set class-wide defaults for EPICS CA communications

        This may be called only before any instances of EpicsSignalBase are
        made.

        This setting applies to the class it is called on and all its
        subclasses. For example,

        >>> EpicsSignalBase.set_defaults(...)

        will apply to ``EpicsSignalRO`` and ``EpicsSignal``, which are both
        subclasses of ``EpicsSignalBase``.

        but

        >>> EpicsSignal.set_defaults(...)

        will not apply to ``EpicsSignalRO``.

        Parameters
        ----------
        auto_monitor: bool, optional
            If ``True``, update cached value from EPICS CA monitor callbacks.
            If ``False``, request new value from EPICS each time get() is called.
        connection_timeout: float, optional
            Time (seconds) allocated for establishing a connection with the
            IOC.
        timeout: float, optional
            Total time budget (seconds) for reading, not including connection time.
        write_timeout: float, optional
            Time (seconds) allocated for writing, not including connection time.
            The write_timeout is very different than the connection and read timeouts
            above. It relates to how long an action takes to complete. Any
            default value we choose here is likely to cause problems---either
            by being too short and giving up too early on a lengthy action or
            being too long and delaying the report of a failure. The default,
            None, waits forever.

        Raises
        ------
        RuntimeError
            If called after :class:`EpicsSignalBase` has been instantiated for
            the first time.
        """
        if EpicsSignalBase.__any_instantiated:
            raise RuntimeError(
                "The method EpicsSignalBase.set_defaults may only "
                "be called before the first instance of EpicsSignalBase is "
                "created. This is to ensure that all instances are created "
                "with the same default settings in place."
            )

        cls.__default_auto_monitor = auto_monitor
        cls.__default_connection_timeout = connection_timeout
        cls.__default_timeout = timeout
        # The write_timeout is very different than the connection and read timeouts
        # above. It relates to how long an action takes to complete. Any
        # default value we choose here is likely to cause problems---either
        # by being too short and giving up too early on a lengthy action or
        # being too long and delaying the report of a failure.
        cls.__default_write_timeout = write_timeout

        # TODO Is there a good reason to prohibit setting these three timeout
        # properties?

    @classmethod
    def set_default_timeout(cls, **kwargs):
        warnings.warn(
            "set_default_timeout() will be removed "
            "in a future release. Use set_defaults() instead."
        )
        cls.set_defaults(**kwargs)

    @property
    def connection_timeout(self):
        return self._connection_timeout

    @property
    def timeout(self):
        return self._timeout

    @property
    def write_timeout(self):
        return self._write_timeout

    def __getnewargs_ex__(self):
        args, kwargs = super().__getnewargs_ex__()
        # 'value' shows up in the EpicsSignal repr, but should not be used to
        # copy the Signal
        kwargs.pop("value", None)
        return (args, kwargs)

    def _initial_metadata_callback(self, pvname, cl_metadata):
        "Control-layer callback: all initial metadata - control and status"
        self._metadata_changed(
            pvname, cl_metadata, require_timestamp=True, update=True, from_monitor=False
        )
        self._received_first_metadata[pvname] = True
        self._set_event_if_ready()

    def _metadata_changed(
        self, pvname, cl_metadata, *, from_monitor, update, require_timestamp=False
    ):
        "Notification: the metadata of a single PV has changed"
        metadata = self._get_metadata_from_kwargs(
            pvname, cl_metadata, require_timestamp=require_timestamp
        )
        if update:
            self._metadata.update(**metadata)
        return metadata

    def _pv_connected(self, pvname, conn, pv):
        "Control-layer callback: PV has [dis]connected"
        if self._destroyed:
            return

        was_connected = self.connected
        if not conn:
            self._signal_is_ready.clear()
            self._metadata["connected"] = False
            self._access_rights_valid[pvname] = False

        self._connection_states[pvname] = conn

        if conn and not self._received_first_metadata[pvname]:
            pv.get_all_metadata_callback(self._initial_metadata_callback, timeout=10)

        self._set_event_if_ready()

        if was_connected and not conn:
            # Send a notification of disconnection
            self._run_metadata_callbacks()

        if self._auto_monitor:
            if getattr(self, "_read_pvname", None) == pvname:
                self._add_callback(pvname, pv, self._read_changed)
            if getattr(self, "_setpoint_pvname", None) == pvname:
                self._add_callback(pvname, pv, self._write_changed)

    def _set_event_if_ready(self):
        """If connected and access rights received, set the "ready" event used
        in wait_for_connection."""
        with self._metadata_lock:
            already_connected = self._metadata["connected"]
            if self._destroyed or already_connected:
                return
            elif not all(
                [
                    *self._connection_states.values(),
                    *self._access_rights_valid.values(),
                    *self._received_first_metadata.values(),
                ]
            ):
                if self._metadata["connected"]:
                    self._metadata["connected"] = False
                    # subs are run in _pv_connected
                return

            self._metadata["connected"] = True
            self._signal_is_ready.set()

        self._run_metadata_callbacks()

    def _pv_access_callback(self, read_access, write_access, pv):
        "Control-layer callback: PV access rights have changed"
        self._access_rights_valid[pv.pvname] = True

    @property
    def as_string(self):
        """Attempt to cast the EPICS PV value to a string by default"""
        return self._string

    @property
    def precision(self):
        """The precision of the read PV, as reported by EPICS"""
        return self._metadata["precision"]

    @property
    def enum_strs(self):
        """List of strings if PV is an enum type"""
        return self._metadata["enum_strs"]

    @property
    def alarm_status(self):
        """PV status"""
        return self._metadata["status"]

    @property
    def alarm_severity(self):
        """PV alarm severity"""
        return self._metadata["severity"]

    def _add_callback(self, pvname, pv, cb):
        with self._metadata_lock:
            if not self._monitors[pvname]:
                mon = pv.add_callback(cb, run_now=pv.connected)
                self._monitors[pvname] = mon

    @doc_annotation_forwarder(Signal)
    def subscribe(self, callback, event_type=None, run=True):
        if event_type is None:
            event_type = self._default_sub
        if event_type == self.SUB_VALUE:
            self._add_callback(self._read_pvname, self._read_pv, self._read_changed)

        return super().subscribe(callback, event_type=event_type, run=run)

    def _ensure_connected(self, *pvs, timeout):
        "Ensure that `pv` is connected, with access/connection callbacks run"
        with self._metadata_lock:
            if self.connected:
                return
            elif self._destroyed:
                raise DestroyedError("Cannot re-use a destroyed Signal")

        for pv in pvs:
            if not pv.wait_for_connection(timeout=timeout):
                raise TimeoutError(
                    f"{pv.pvname} could not connect within "
                    f"{float(timeout):.3}-second timeout."
                )

        for pv in pvs:
            if not self._received_first_metadata[pv.pvname]:
                # Utility threads can get backed up in cases of PV connection
                # storms.  Since the user is specifically blocking on this PV,
                # make it a priority and perform the request in the current
                # thread.
                md = pv.get_all_metadata_blocking(timeout=timeout)
                self._initial_metadata_callback(pv.pvname, md)

        # Ensure callbacks are run prior to returning, as
        # @raise_if_disconnected can cause issues otherwise.
        if not self._signal_is_ready.wait(timeout):
            raise TimeoutError(
                f"Control layer {self.cl.name} failed to send connection and "
                f"access rights information within {float(timeout):.1f} sec"
            )

    def wait_for_connection(self, timeout=DEFAULT_CONNECTION_TIMEOUT):
        """Wait for the underlying signals to initialize or connect"""
        if timeout is DEFAULT_CONNECTION_TIMEOUT:
            timeout = self.connection_timeout
        try:
            self._ensure_connected(self._read_pv, timeout=timeout)
        except TimeoutError:
            if self._destroyed:
                raise DestroyedError("Signal has been destroyed")
            raise

    @property
    def timestamp(self):
        """Timestamp of readback PV, according to EPICS"""
        return self._metadata["timestamp"]

    @property
    def pvname(self):
        """The readback PV name"""
        return self._read_pvname

    def _repr_info(self):
        "Yields pairs of (key, value) to generate the Signal repr"
        yield ("read_pv", self.pvname)
        yield from super()._repr_info()
        yield ("auto_monitor", self._auto_monitor)
        yield ("string", self._string)

    @property
    def limits(self):
        """The PV control limits (low, high), such that low <= value <= high"""
        # This overrides the base Signal limits
        return (self._metadata["lower_ctrl_limit"], self._metadata["upper_ctrl_limit"])

    def _get_with_timeout(
        self, pv, timeout, connection_timeout, as_string, form, use_monitor
    ):
        """
        Utility method implementing a retry loop for get and get_setpoint

        Returns info from pv.read_with_metadata(...) or raises TimeoutError
        """
        # Fall back to instance default if no value is given.
        if timeout is DEFAULT_TIMEOUT:
            timeout = self.timeout
        if connection_timeout is DEFAULT_CONNECTION_TIMEOUT:
            connection_timeout = self.connection_timeout

        if connection_timeout is not None:
            connection_timeout = float(connection_timeout)

        try:
            self.wait_for_connection(timeout=connection_timeout)
        except TimeoutError as err:
            raise ConnectionTimeoutError(
                f"Failed to connect to {pv.pvname} "
                f"within {connection_timeout:.2f} sec"
            ) from err
        # Pyepics returns None when a read request times out.  Raise a
        # TimeoutError on its behalf.
        self.control_layer_log.debug(
            "pv[%s].get_with_metadata(as_string=%s, form=%s, timeout=%s)",
            pv.pvname,
            as_string,
            form,
            timeout,
        )
        info = pv.get_with_metadata(
            as_string=as_string, form=form, timeout=timeout, use_monitor=use_monitor
        )
        self.control_layer_log.debug(
            "pv[%s].get_with_metadata(...) returned", pv.pvname
        )

        if info is None:
            raise ReadTimeoutError(
                f"Failed to read {pv.pvname} " f"within {timeout:.2f} sec"
            )

        return info

    def get(
        self,
        *,
        as_string=None,
        timeout=DEFAULT_TIMEOUT,
        connection_timeout=DEFAULT_CONNECTION_TIMEOUT,
        form="time",
        use_monitor=None,
        **kwargs,
    ):
        """Get the readback value through an explicit call to EPICS.

        Parameters
        ----------
        count : int, optional
            Explicitly limit count for array data
        as_string : bool, optional
            Get a string representation of the value, defaults to as_string
            from this signal, optional
        as_numpy : bool
            Use numpy array as the return type for array data.
        timeout : float, optional
            maximum time to wait for value to be received.
            (default = 0.5 + log10(count) seconds)
        use_monitor : bool, optional
            to use value from latest monitor callback or to make an
            explicit CA call for the value. (default: True)
        connection_timeout : float, optional
            If not already connected, allow up to `connection_timeout` seconds
            for the connection to complete.
        form : {'time', 'ctrl'}
            PV form to request

        """
        if kwargs:
            warnings.warn(
                "Signal.get no longer takes keyword arguments; "
                "These are ignored and will be deprecated.",
                DeprecationWarning,
            )
        if as_string is None:
            as_string = self._string

        if use_monitor is None:
            use_monitor = self._auto_monitor

        info = self._get_with_timeout(
            self._read_pv, timeout, connection_timeout, as_string, form, use_monitor
        )

        value = info.pop("value")
        if as_string:
            value = waveform_to_string(value)

        has_monitor = self._monitors[self.pvname] is not None
        if form != "time" or not has_monitor:
            # Different form such as 'ctrl' holds additional data not available
            # through the DBR_TIME channel access monitor
            self._metadata_changed(
                self.pvname,
                info,
                update=True,
                require_timestamp=True,
                from_monitor=False,
            )

        if not has_monitor:
            # No monitor - readback can only be updated here
            self._readback = value
        return self._fix_type(value)

    def _fix_type(self, value):
        "Cast the given value according to the data type of this EpicsSignal"
        if self._string:
            value = waveform_to_string(value)

        return value

    def _get_metadata_from_kwargs(
        self, pvname, cl_metadata, *, require_timestamp=False
    ):
        "Metadata from the control layer -> metadata for this Signal"

        def fix_value(fixer_function, value):
            return (
                fixer_function(value)
                if fixer_function is not None and value is not None
                else value
            )

        metadata = {
            md_key: fix_value(fixer_function, cl_metadata[cl_key])
            for cl_key, (md_key, fixer_function) in self._metadata_key_map[
                pvname
            ].items()
            if cl_metadata.get(cl_key, None) is not None
        }

        if require_timestamp and metadata.get("timestamp", None) is None:
            metadata["timestamp"] = time.time()
        return metadata

    def _read_changed(self, value=None, **kwargs):
        "CA monitor callback indicating that the read value has changed"
        metadata = self._metadata_changed(
            self.pvname, kwargs, update=False, require_timestamp=True, from_monitor=True
        )

        if self._string and "char_value" in kwargs:
            value = kwargs["char_value"]

        # super().put updates self._readback and runs SUB_VALUE
        super().put(
            value=value,
            timestamp=metadata.pop("timestamp"),
            metadata=metadata,
            force=True,
        )

    def describe(self):
        """Return the description as a dictionary

        Returns
        -------
        dict
            Dictionary of name and formatted description string
        """
        if self._readback is DEFAULT_EPICSSIGNAL_VALUE:
            val = self.get()
        else:
            val = self._readback
        lower_ctrl_limit, upper_ctrl_limit = self.limits
        desc = dict(
            source="PV:{}".format(self._read_pvname),
            dtype=data_type(val),
            shape=data_shape(val),
            units=self._metadata["units"],
            lower_ctrl_limit=lower_ctrl_limit,
            upper_ctrl_limit=upper_ctrl_limit,
        )

        if self.precision is not None:
            desc["precision"] = self.precision

        if self.enum_strs is not None:
            desc["enum_strs"] = tuple(self.enum_strs)

        return {self.name: desc}


class EpicsSignalRO(EpicsSignalBase):
    """A read-only EpicsSignal -- that is, one with no `write_pv`

    Keyword arguments are passed on to the base class (Signal) initializer

    Parameters
    ----------
    read_pv : str
        The PV to read from
    limits : bool, optional
        Check limits prior to writing value
    auto_monitor : bool, optional
        Use automonitor with epics.PV
    name : str, optional
        Name of signal.  If not given defaults to read_pv
    timeout : float or None, optional
        The timeout for serving a read request on a connected channel. This is
        only applied if the PV is connected within connection_timeout (below).

        The default value DEFAULT_TIMEOUT means, "Fall back to class-wide
        default." See EpicsSignalBase.set_defaults to configure class
        defaults.

        Explicitly passing None means, "Wait forever."
    write_timeout : float or None, optional
        The timeout for a reply when put completion is used. This is
        only applied if the PV is connected within connection_timeout (below).

        This is very different than the connection and read timeouts
        above. It relates to how long an action takes to complete, such motor
        motion or data acquisition. Any default value we choose here is likely
        to cause problems---either by being too short and giving up too early
        on a lengthy action or being too long and delaying the report of a
        failure. A finite value can be injected here or, perhaps more usefully,
        via `set` at the Device level, where a context-appropriate value can be
        chosen.
    connection_timeout : float or None, optional
        Timeout for connection. This includes the time to search and establish
        a channel.

        The default value DEFAULT_CONNECTION_TIMEOUT means, "Fall back to
        class-wide default." See EpicsSignalBase.set_defaults to
        configure class defaults.

        Explicitly passing None means, "Wait forever."
    """

    def __init__(self, read_pv, *, string=False, name=None, **kwargs):
        super().__init__(read_pv, string=string, name=name, **kwargs)
        self._metadata["write_access"] = False

    def put(self, *args, **kwargs):
        "Disabled for a read-only signal"
        raise ReadOnlyError("Cannot write to read-only EpicsSignal")

    def set(self, *args, **kwargs):
        "Disabled for a read-only signal"
        raise ReadOnlyError("Read-only signals cannot be set")

    def _pv_access_callback(self, read_access, write_access, pv):
        "Control-layer callback: read PV access rights have changed"
        # Tweak write access here - this is a read-only signal!
        if self._destroyed:
            return

        self._metadata.update(
            read_access=read_access,
            write_access=False,
        )

        was_connected = self.connected
        super()._pv_access_callback(read_access, write_access, pv)
        self._set_event_if_ready()

        if was_connected:
            # _set_event_if_ready, above, will run metadata callbacks
            self._run_metadata_callbacks()


class EpicsSignalNoValidation(EpicsSignalBase):
    """An EpicsSignal that does not verify values on set.
    This signal does support readback, but does not guarantee that
    the readback will match the set value.

    Keyword arguments are passed on to the base class (Signal) initializer

    Parameters
    ----------
    write_pv : str
        The PV to write to
    name : str, optional
        Name of signal.  If not given defaults to write_pv
    write_timeout : float or None, optional
        The timeout for a reply when put completion is used. This is
        only applied if the PV is connected within connection_timeout (below).

        This is very different than the connection and read timeouts
        above. It relates to how long an action takes to complete, such motor
        motion or data acquisition. Any default value we choose here is likely
        to cause problems---either by being too short and giving up too early
        on a lengthy action or being too long and delaying the report of a
        failure. A finite value can be injected here or, perhaps more usefully,
        via `set` at the Device level, where a context-appropriate value can be
        chosen.
    connection_timeout : float or None, optional
        Timeout for connection. This includes the time to search and establish
        a channel.

        The default value DEFAULT_CONNECTION_TIMEOUT means, "Fall back to
        class-wide default." See EpicsSignalBase.set_defaults to
        configure class defaults.

        Explicitly passing None means, "Wait forever."
    """

    def set(self, value, *args, **kwargs):
        """
        Set the value of this signal, and return a completed Status
        object, bypassing any readback verification

        Returns
        -------
        st : Status
            This status object will be finished
        """
        self.put(value)

        st = Status(self)
        st.set_finished()
        st.wait()
        return st


class EpicsSignal(EpicsSignalBase):
    """An EPICS signal, comprised of either one or two EPICS PVs

    Keyword arguments are passed on to the base class (Signal) initializer

    Parameters
    ----------
    read_pv : str
        The PV to read from
    write_pv : str, optional
        The PV to write to if different from the read PV
    limits : bool, optional
        Check limits prior to writing value
    auto_monitor : bool, optional
        Use automonitor with epics.PV
    name : str, optional
        Name of signal.  If not given defaults to read_pv
    put_complete : bool, optional
        Use put completion when writing the value
    tolerance : any, optional
        The absolute tolerance associated with the value.
        If specified, this overrides any precision information calculated from
        the write PV
    rtolerance : any, optional
        The relative tolerance associated with the value
    timeout : float or None, optional
        The timeout for serving a read request on a connected channel. This is
        only applied if the PV is connected within connection_timeout (below).

        The default value DEFAULT_TIMEOUT means, "Fall back to class-wide
        default." See EpicsSignalBase.set_defaults to configure class
        defaults.

        Explicitly passing None means, "Wait forever."
    write_timeout : float or None, optional
        The timeout for a reply when put completion is used. This is
        only applied if the PV is connected within connection_timeout (below).

        This is very different than the connection and read timeouts
        above. It relates to how long an action takes to complete, such motor
        motion or data acquisition. Any default value we choose here is likely
        to cause problems---either by being too short and giving up too early
        on a lengthy action or being too long and delaying the report of a
        failure. A finite value can be injected here or, perhaps more usefully,
        via `set` at the Device level, where a context-appropriate value can be
        chosen.
    connection_timeout : float or None, optional
        Timeout for connection. This includes the time to search and establish
        a channel.

        The default value DEFAULT_CONNECTION_TIMEOUT means, "Fall back to
        class-wide default." See EpicsSignalBase.set_defaults to
        configure class defaults.

        Explicitly passing None means, "Wait forever."
    """

    SUB_SETPOINT = "setpoint"
    SUB_SETPOINT_META = "setpoint_meta"

    _write_pv_metadata_key_map = dict(
        status=("setpoint_status", AlarmStatus),
        severity=("setpoint_severity", AlarmSeverity),
        precision=("setpoint_precision", None),
        timestamp=("setpoint_timestamp", None),
        # Override the readback ones, as we write to the setpoint:
        lower_ctrl_limit=("lower_ctrl_limit", None),
        upper_ctrl_limit=("upper_ctrl_limit", None),
    )

    _metadata_keys = EpicsSignalBase._metadata_keys + (
        "setpoint_status",
        "setpoint_severity",
        "setpoint_precision",
        "setpoint_timestamp",
    )

    def __init__(
        self,
        read_pv,
        write_pv=None,
        *,
        put_complete=False,
        string=False,
        limits=False,
        name=None,
        **kwargs,
    ):

        self._write_pv = None
        self._use_limits = bool(limits)
        self._put_complete = put_complete
        self._setpoint = None

        metadata = dict(
            setpoint_timestamp=None,
            setpoint_status=None,
            setpoint_severity=None,
            lower_ctrl_limit=None,
            upper_ctrl_limit=None,
        )

        if write_pv is None:
            write_pv = read_pv

        self._setpoint_pvname = write_pv

        super().__init__(
            read_pv,
            string=string,
            name=name,
            metadata=metadata,
            all_pvs={read_pv, write_pv},
            **kwargs,
        )

        if read_pv == write_pv:
            self._write_pv = self._read_pv
        else:
            validate_pv_name(write_pv)
            self._metadata_key_map = {
                write_pv: self._write_pv_metadata_key_map,
                read_pv: {
                    key: value
                    for key, value in self._metadata_key_map[read_pv].items()
                    if key not in ("lower_ctrl_limit", "upper_ctrl_limit")
                },
            }

            self._write_pv = self.cl.get_pv(
                write_pv,
                auto_monitor=self._auto_monitor,
                connection_callback=self._pv_connected,
                access_callback=self._pv_access_callback,
            )

        self._write_pv._reference_count += 1

        # NOTE: after this point, write_pv can either be:
        #  (1) the same as read_pv
        #  (2) a completely separate PV instance
        # It will not be None, until destroy() is called.

        def finalize(write_pv, cl):
            cl.release_pvs(write_pv)

        self._write_pv_finalizer = weakref.finalize(
            self, finalize, self._write_pv, self.cl
        )

    def destroy(self):
        super().destroy()
        self._write_pv_finalizer()

    @doc_annotation_forwarder(EpicsSignalBase)
    def subscribe(self, callback, event_type=None, run=True):
        if event_type is None:
            event_type = self._default_sub

        if event_type == self.SUB_SETPOINT:
            self._add_callback(
                self._setpoint_pvname, self._write_pv, self._write_changed
            )

        return super().subscribe(callback, event_type=event_type, run=run)

    def wait_for_connection(self, timeout=DEFAULT_CONNECTION_TIMEOUT):
        """Wait for the underlying signals to initialize or connect"""
        if timeout is DEFAULT_CONNECTION_TIMEOUT:
            timeout = self.connection_timeout
        self._ensure_connected(self._read_pv, self._write_pv, timeout=timeout)

    @property
    def tolerance(self):
        """The tolerance of the write PV, as reported by EPICS

        Can be overidden by the user at the EpicsSignal level.

        Returns
        -------
        tolerance : float or None
        Using the write PV's precision:
            If precision == 0, tolerance will be None
            If precision > 0, calculated to be 10**(-precision)
        """
        # NOTE: overrides Signal.tolerance property
        if self._tolerance is not None:
            return self._tolerance

        precision = self.precision
        if precision == 0 or precision is None:
            return None

        return 10.0 ** (-precision)

    @tolerance.setter
    def tolerance(self, tolerance):
        self._tolerance = tolerance

    @property
    def setpoint_ts(self):
        """Timestamp of setpoint PV, according to EPICS"""
        return self._metadata["setpoint_timestamp"]

    @property
    def setpoint_pvname(self):
        """The setpoint PV name"""
        return self._setpoint_pvname

    @property
    def setpoint_alarm_status(self):
        """Setpoint PV status"""
        return self._metadata["setpoint_status"]

    @property
    def setpoint_alarm_severity(self):
        """Setpoint PV alarm severity"""
        return self._metadata["setpoint_severity"]

    def _repr_info(self):
        "Yields pairs of (key, value) to generate the Signal repr"
        yield from super()._repr_info()
        yield ("write_pv", self._setpoint_pvname)
        yield ("limits", self._use_limits)
        yield ("put_complete", self._put_complete)

    def check_value(self, value):
        """Check if the value is within the setpoint PV's control limits

        Raises
        ------
        ValueError
        """
        super().check_value(value)

        if value is None:
            raise ValueError("Cannot write None to epics PVs")
        if not self._use_limits:
            return

        low_limit, high_limit = self.limits
        if low_limit >= high_limit:
            return

        if not (low_limit <= value <= high_limit):
            raise LimitError(
                "{}: value {} outside of range: [{}, {}]".format(
                    self.name, value, low_limit, high_limit
                )
            )

    def get_setpoint(
        self,
        *,
        as_string=None,
        timeout=DEFAULT_TIMEOUT,
        connection_timeout=DEFAULT_CONNECTION_TIMEOUT,
        use_monitor=None,
        form="time",
        **kwargs,
    ):
        """Get the setpoint value (if setpoint PV and readback PV differ)

        Parameters
        ----------
        count : int, optional
            Explicitly limit count for array data
        as_string : bool, optional
            Get a string representation of the value, defaults to as_string
            from this signal, optional
        as_numpy : bool
            Use numpy array as the return type for array data.
        timeout : float, optional
            maximum time to wait for value to be received.
            (default = 0.5 + log10(count) seconds)
        use_monitor : bool, optional
            to use value from latest monitor callback or to make an
            explicit CA call for the value. (default: True)
        connection_timeout : float, optional
            If not already connected, allow up to `connection_timeout` seconds
            for the connection to complete.
        form : {'time', 'ctrl'}
            PV form to request
        """
        if kwargs:
            warnings.warn(
                "Signal.get_setpoint no longer takes keyword arguments; "
                "These are ignored and will be deprecated.",
                DeprecationWarning,
            )
        if as_string is None:
            as_string = self._string

        if use_monitor is None:
            use_monitor = self._auto_monitor

        info = self._get_with_timeout(
            self._write_pv, timeout, connection_timeout, as_string, form, use_monitor
        )

        value = info.pop("value")
        if as_string:
            value = waveform_to_string(value)

        has_monitor = self._monitors[self.setpoint_pvname] is not None
        if form != "time" or not has_monitor:
            # Different form such as 'ctrl' holds additional data not available
            # through the DBR_TIME channel access monitor
            self._metadata_changed(
                self.setpoint_pvname,
                info,
                update=True,
                from_monitor=False,
                require_timestamp=True,
            )

        if not has_monitor:
            # No monitor - setpoint can only be updated here
            self._setpoint = self._fix_type(value)
        return self._fix_type(value)

    def _pv_access_callback(self, read_access, write_access, pv):
        "Control-layer callback: PV access rights have changed"
        if self._destroyed:
            return

        md_update = {}
        if pv.pvname == self.pvname:
            md_update["read_access"] = read_access

        if pv.pvname == self.setpoint_pvname:
            md_update["write_access"] = write_access

        if md_update:
            self._metadata.update(**md_update)

        if self.connected:
            self._run_metadata_callbacks()

        super()._pv_access_callback(read_access, write_access, pv)
        self._set_event_if_ready()

    def _metadata_changed(
        self, pvname, cl_metadata, *, from_monitor, update, require_timestamp=False
    ):
        "Metadata for one PV has changed"
        metadata = super()._metadata_changed(
            pvname,
            cl_metadata,
            from_monitor=from_monitor,
            update=update,
            require_timestamp=require_timestamp,
        )

        if all(
            (
                from_monitor,
                self.setpoint_pvname != self.pvname,
                self.setpoint_pvname == pvname,
            )
        ):
            # Setpoint has its own metadata
            self._metadata_thread_ctx.run(
                self._run_subs,
                sub_type=self.SUB_SETPOINT_META,
                timestamp=self._metadata["setpoint_timestamp"],
                status=self._metadata["setpoint_status"],
                severity=self._metadata["setpoint_severity"],
                precision=self._metadata["setpoint_precision"],
                lower_ctrl_limit=self._metadata["lower_ctrl_limit"],
                upper_ctrl_limit=self._metadata["upper_ctrl_limit"],
                units=self._metadata["units"],
            )
        return metadata

    def _write_changed(self, value=None, timestamp=None, **kwargs):
        "CA monitor: callback indicating the setpoint PV value has changed"
        if timestamp is None:
            timestamp = time.time()

        self._metadata_changed(
            self.setpoint_pvname,
            kwargs,
            require_timestamp=True,
            from_monitor=True,
            update=True,
        )

        old_value = self._setpoint
        self._setpoint = self._fix_type(value)

        self._run_subs(
            sub_type=self.SUB_SETPOINT,
            old_value=old_value,
            value=value,
            timestamp=self._metadata["setpoint_timestamp"],
            status=self._metadata["setpoint_status"],
            severity=self._metadata["setpoint_severity"],
        )

    def put(
        self,
        value,
        force=False,
        connection_timeout=DEFAULT_CONNECTION_TIMEOUT,
        callback=None,
        use_complete=None,
        timeout=DEFAULT_WRITE_TIMEOUT,
        **kwargs,
    ):
        """
        Using channel access, set the write PV to `value`.

        Keyword arguments are passed on to callbacks

        Parameters
        ----------
        value : any
            The value to set
        force : bool, optional
            Skip checking the value in Python first
        connection_timeout : float, optional
            If not already connected, allow up to `connection_timeout` seconds
            for the connection to complete.
        use_complete : bool, optional
            Override put completion settings
        callback : callable
            Callback for when the put has completed
        timeout : float, optional
            Timeout before assuming that put has failed. (Only relevant if
            put completion is used.)
        """
        if not force:
            self.check_value(value)

        if connection_timeout is DEFAULT_CONNECTION_TIMEOUT:
            connection_timeout = self.connection_timeout
        if timeout is DEFAULT_WRITE_TIMEOUT:
            timeout = self.write_timeout
        self.wait_for_connection(timeout=connection_timeout)
        if use_complete is None:
            use_complete = self._put_complete

        if not self.write_access:
            raise ReadOnlyError("No write access to underlying EPICS PV")

        self.control_layer_log.debug(
            "_write_pv.put(value=%s, use_complete=%s, callback=%s, kwargs=%s)",
            value,
            use_complete,
            callback,
            kwargs,
        )
        self._write_pv.put(
            value,
            use_complete=use_complete,
            callback=callback,
            timeout=timeout,
            **kwargs,
        )

        old_value = self._setpoint
        self._setpoint = value

        if self.pvname == self.setpoint_pvname:
            timestamp = time.time()
            super().put(value, timestamp=timestamp, force=True)
            self._run_subs(
                sub_type=self.SUB_SETPOINT,
                old_value=old_value,
                value=value,
                timestamp=timestamp,
            )

    def set(self, value, *, timeout=DEFAULT_WRITE_TIMEOUT, settle_time=None):
        """
        Set the value of the Signal and return a Status object.

        If put completion is used for this EpicsSignal, the status object will
        complete once EPICS reports the put has completed.

        Otherwise the readback will be polled until equal to the set point (as
        in `Signal.set`)

        Parameters
        ----------
        value : any
        timeout : float, optional
            Maximum time to wait.
        settle_time: float, optional
            Delay after the set() has completed to indicate completion
            to the caller

        Returns
        -------
        st : Status

        See Also
        --------
        Signal.set

        """
        if timeout is DEFAULT_WRITE_TIMEOUT:
            timeout = self.write_timeout

        if not self._put_complete:
            return super().set(value, timeout=timeout, settle_time=settle_time)

        # using put completion:
        # timeout and settle time is handled by the status object.
        st = Status(self, timeout=timeout, settle_time=settle_time)

        def put_callback(**kwargs):
            st._finished(success=True)

        self.put(value, use_complete=True, callback=put_callback)
        return st

    @property
    def setpoint(self):
        """The setpoint PV value"""
        if self._setpoint is not None:
            return self._setpoint

        return self.get_setpoint()

    @setpoint.setter
    def setpoint(self, value):
        warnings.warn(
            "Setting EpicsSignal.setpoint is deprecated and " "will be removed"
        )
        self.put(value)

    @property
    def put_complete(self):
        "Use put completion when writing the value"
        return self._put_complete

    @put_complete.setter
    def put_complete(self, value):
        self._put_complete = bool(value)

    @property
    def use_limits(self):
        "Check value against limits prior to sending to EPICS"
        return self._use_limits

    @use_limits.setter
    def use_limits(self, value):
        self._use_limits = bool(value)


class AttributeSignal(Signal):
    """Signal derived from a Python object instance's attribute

    Parameters
    ----------
    attr : str
        The dotted attribute name, relative to this signal's parent.
    name : str, optional
        The signal name
    parent : Device, optional
        The parent device instance
    read_access : bool, optional
        Allow read access to the attribute
    write_access : bool, optional
        Allow write access to the attribute
    """

    def __init__(self, attr, *, name=None, parent=None, write_access=True, **kwargs):
        super().__init__(name=name, parent=parent, **kwargs)

        if "." in attr:
            self.attr_base, self.attr = attr.rsplit(".", 1)
        else:
            self.attr_base, self.attr = None, attr

        self._metadata.update(
            read_access=True,
            write_access=write_access,
        )

    @property
    def full_attr(self):
        """The full attribute name"""
        if not self.attr_base:
            return self.attr
        else:
            return ".".join((self.attr_base, self.attr))

    @property
    def base(self):
        """The parent instance which has the final attribute"""
        if self.attr_base is None:
            return self.parent

        obj = self.parent
        for i, part in enumerate(self.attr_base.split(".")):
            try:
                obj = getattr(obj, part)
            except AttributeError as ex:
                raise AttributeError("{}.{} ({})".format(obj.name, part, ex))

        return obj

    def get(self, **kwargs):
        "Get the value from the associated attribute"
        self._readback = getattr(self.base, self.attr)
        return self._readback

    def put(self, value, **kwargs):
        "Write to the associated attribute"
        if not self.write_access:
            raise ReadOnlyError("AttributeSignal is marked as read-only")

        old_value = self.get()
        setattr(self.base, self.attr, value)
        self._run_subs(
            sub_type=self.SUB_VALUE,
            old_value=old_value,
            value=value,
            timestamp=time.time(),
        )

    def describe(self):
        value = self.get()
        desc = {
            "source": "PY:{}.{}".format(self.parent.name, self.full_attr),
            "dtype": data_type(value),
            "shape": data_shape(value),
        }
        return {self.name: desc}


class ArrayAttributeSignal(AttributeSignal):
    """An AttributeSignal which is cast to an ndarray on get

    This is used where data_type and data_shape may otherwise fail to determine
    how to store the data into metadatastore.
    """

    def get(self, **kwargs):
        return np.asarray(super().get(**kwargs))



from collections import deque
from logging import LoggerAdapter
import threading
import time
from warnings import warn

import numpy as np

from .log import logger
from .utils import (
    adapt_old_callback_signature,
    InvalidState,
    UnknownStatusFailure,
    StatusTimeoutError,
    WaitTimeoutError,
)


class UseNewProperty(RuntimeError):
    ...


class StatusBase:
    """
    Track the status of a potentially-lengthy action like moving or triggering.

    Parameters
    ----------
    timeout: float, optional
        The amount of time to wait before marking the Status as failed.  If
        ``None`` (default) wait forever. It is strongly encouraged to set a
        finite timeout.  If settle_time below is set, that time is added to the
        effective timeout.
    settle_time: float, optional
        The amount of time to wait between the caller specifying that the
        status has completed to running callbacks. Default is 0.


    Notes
    -----

    Theory of operation:

    This employs two ``threading.Event`` objects, one thread the runs for
    (timeout + settle_time) seconds, and one thread that runs for
    settle_time seconds (if settle_time is nonzero).

    At __init__ time, a *timeout* and *settle_time* are specified. A thread
    is started, on which user callbacks, registered after __init__ time via
    :meth:`add_callback`, will eventually be run. The thread waits on an
    Event be set or (timeout + settle_time) seconds to pass, whichever
    happens first.

    If (timeout + settle_time) expires and the Event has not
    been set, an internal Exception is set to ``StatusTimeoutError``, and a
    second Event is set, marking the Status as done and failed. The
    callbacks are run.

    If a callback is registered after the Status is done, it will be run
    immediately.

    If the first Event is set before (timeout + settle_time) expires,
    then the second Event is set and no internal Exception is set, marking
    the Status as done and successful. The callbacks are run.

    There are two methods that directly set the first Event. One,
    :meth:set_exception, sets it directly after setting the internal
    Exception.  The other, :meth:`set_finished`, starts a
    ``threading.Timer`` that will set it after a delay (the settle_time).
    One of these methods may be called, and at most once. If one is called
    twice or if both are called, ``InvalidState`` is raised. If they are
    called too late to prevent a ``StatusTimeoutError``, they are ignored
    but one call is still allowed. Thus, an external callback, e.g. pyepics,
    may reports success or failure after the Status object has expired, but
    to no effect because the callbacks have already been called and the
    program has moved on.

    """

    def __init__(self, *, timeout=None, settle_time=0, done=None, success=None):
        super().__init__()
        self._tname = None
        self._lock = threading.RLock()
        self._event = threading.Event()  # state associated with done-ness
        self._settled_event = threading.Event()
        # "Externally initiated" means set_finished() or set_exception(exc) was
        # called, as opposed to completion via an internal timeout.
        self._externally_initiated_completion_lock = threading.Lock()
        self._externally_initiated_completion = False
        self._callbacks = deque()
        self._exception = None

        self.log = LoggerAdapter(logger=logger, extra={"status": self})

        if settle_time is None:
            settle_time = 0.0

        self._settle_time = float(settle_time)

        if timeout is not None:
            timeout = float(timeout)
        self._timeout = timeout

        # We cannot know that we are successful if we are not done.
        if success and not done:
            raise ValueError("Cannot initialize with done=False but success=True.")
        if done is not None or success is not None:
            warn(
                "The 'done' and 'success' parameters will be removed in a "
                "future release. Use the methods set_finished() or "
                "set_exception(exc) to mark success or failure, respectively, "
                "after the Status has been instantiated.",
                DeprecationWarning,
            )

        self._callback_thread = threading.Thread(
            target=self._run_callbacks, daemon=True, name=self._tname
        )
        self._callback_thread.start()

        if done:
            if success:
                self.set_finished()
            else:
                exc = UnknownStatusFailure(
                    f"The status {self!r} has failed. To obtain more specific, "
                    "helpful errors in the future, update the Device to use "
                    "set_exception(...) instead of setting success=False "
                    "at __init__ time."
                )
                self.set_exception(exc)

    @property
    def timeout(self):
        """
        The timeout for this action.

        This is set when the Status is created, and it cannot be changed.
        """
        return self._timeout

    @property
    def settle_time(self):
        """
        A delay between when :meth:`set_finished` is when the Status is done.

        This is set when the Status is created, and it cannot be changed.
        """
        return self._settle_time

    @property
    def done(self):
        """
        Boolean indicating whether associated operation has completed.

        This is set to True at __init__ time or by calling
        :meth:`set_finished`, :meth:`set_exception`, or (deprecated)
        :meth:`_finished`. Once True, it can never become False.
        """
        return self._event.is_set()

    @done.setter
    def done(self, value):
        # For now, allow this setter to work only if it has no effect.
        # In a future release, make this property not settable.
        if bool(self._event.is_set()) != bool(value):
            raise RuntimeError(
                "The done-ness of a status object cannot be changed by "
                "setting its `done` attribute directly. Call `set_finished()` "
                "or `set_exception(exc)."
            )
        warn(
            "Do not set the `done` attribute of a status object directly. "
            "It should only be set indirectly by calling `set_finished()` "
            "or `set_exception(exc)`. "
            "Direct setting was never intended to be supported and it will be "
            "disallowed in a future release of ophyd, causing this code path "
            "to fail.",
            UserWarning,
        )

    @property
    def success(self):
        """
        Boolean indicating whether associated operation has completed.

        This is set to True at __init__ time or by calling
        :meth:`set_finished`, :meth:`set_exception`, or (deprecated)
        :meth:`_finished`. Once True, it can never become False.
        """
        return self.done and self._exception is None

    @success.setter
    def success(self, value):
        # For now, allow this setter to work only if it has no effect.
        # In a future release, make this property not settable.
        if bool(self.success) != bool(value):
            raise RuntimeError(
                "The success state of a status object cannot be changed by "
                "setting its `success` attribute directly. Call "
                "`set_finished()` or `set_exception(exc)`."
            )
        warn(
            "Do not set the `success` attribute of a status object directly. "
            "It should only be set indirectly by calling `set_finished()` "
            "or `set_exception(exc)`. "
            "Direct setting was never intended to be supported and it will be "
            "disallowed in a future release of ophyd, causing this code path "
            "to fail.",
            UserWarning,
        )

    def _handle_failure(self):
        pass

    def _settled(self):
        """Hook for when status has completed and settled"""
        pass

    def _run_callbacks(self):
        """
        Set the Event and run the callbacks.
        """
        if self.timeout is None:
            timeout = None
        else:
            timeout = self.timeout + self.settle_time
        if not self._settled_event.wait(timeout):
            # We have timed out. It's possible that set_finished() has already
            # been called but we got here before the settle_time timer expired.
            # And it's possible that in this space be between the above
            # statement timing out grabbing the lock just below,
            # set_exception(exc) has been called. Both of these possibilties
            # are accounted for.
            self.log.warning("%r has timed out", self)
            with self._externally_initiated_completion_lock:
                # Set the exception and mark the Status as done, unless
                # set_exception(exc) was called externally before we grabbed
                # the lock.
                if self._exception is None:
                    exc = StatusTimeoutError(
                        f"Status {self!r} failed to complete in specified timeout."
                    )
                    self._exception = exc
        # Mark this as "settled".
        try:
            self._settled()
        except Exception:
            # No alternative but to log this. We can't supersede set_exception,
            # and we have to continue and run the callbacks.
            self.log.exception("%r encountered error during _settled()", self)
        # Now we know whether or not we have succeed or failed, either by
        # timeout above or by set_exception(exc), so we can set the Event that
        # will mark this Status as done.
        with self._lock:
            self._event.set()
        if self._exception is not None:
            try:
                self._handle_failure()
            except Exception:
                self.log.exception(
                    "%r encountered an error during _handle_failure()", self
                )
        # The callbacks have access to self, from which they can distinguish
        # success or failure.
        for cb in self._callbacks:
            try:
                cb(self)
            except Exception:
                self.log.exception(
                    "An error was raised on a background thread while "
                    "running the callback %r(%r).",
                    cb,
                    self,
                )
        self._callbacks.clear()

    def set_exception(self, exc):
        """
        Mark as finished but failed with the given Exception.

        This method should generally not be called by the *recipient* of this
        Status object, but only by the object that created and returned it.

        Parameters
        ----------
        exc: Exception
        """
        # Since we rely on this being raise-able later, check proactively to
        # avoid potentially very confusing failures.
        if not (
            isinstance(exc, Exception)
            or isinstance(exc, type)
            and issubclass(exc, Exception)
        ):
            # Note that Python allows `raise Exception` or raise Exception()`
            # so we allow a class or an instance here too.
            raise ValueError(f"Expected an Exception, got {exc!r}")

        # Ban certain Timeout subclasses that have special significance. This
        # would probably never come up except due to some rare user error, but
        # if it did it could be very confusing indeed!
        for exc_class in (StatusTimeoutError, WaitTimeoutError):
            if (
                isinstance(exc, exc_class)
                or isinstance(exc, type)
                and issubclass(exc, exc_class)
            ):
                raise ValueError(
                    f"{exc_class} has special significance and cannot be set "
                    "as the exception. Use a plain TimeoutError or some other "
                    "subclass thereof."
                )

        with self._externally_initiated_completion_lock:
            if self._externally_initiated_completion:
                raise InvalidState(
                    "Either set_finished() or set_exception() has "
                    f"already been called on {self!r}"
                )
            self._externally_initiated_completion = True
            if isinstance(self._exception, StatusTimeoutError):
                # We have already timed out.
                return
            self._exception = exc
            self._settled_event.set()

    def set_finished(self):
        """
        Mark as finished successfully.

        This method should generally not be called by the *recipient* of this
        Status object, but only by the object that created and returned it.
        """
        with self._externally_initiated_completion_lock:
            if self._externally_initiated_completion:
                raise InvalidState(
                    "Either set_finished() or set_exception() has "
                    f"already been called on {self!r}"
                )
            self._externally_initiated_completion = True
        # Note that in either case, the callbacks themselves are run from the
        # same thread. This just sets an Event, either from this thread (the
        # one calling set_finished) or the thread created below.
        if self.settle_time > 0:
            threading.Timer(self.settle_time, self._settled_event.set).start()
        else:
            self._settled_event.set()

    def _finished(self, success=True, **kwargs):
        """
        Inform the status object that it is done and if it succeeded.

        This method is deprecated. Please use :meth:`set_finished` or
        :meth:`set_exception`.

        .. warning::

           kwargs are not used, but are accepted because pyepics gives
           in a bunch of kwargs that we don't care about.  This allows
           the status object to be handed directly to pyepics (but
           this is probably a bad idea for other reason.

           This may be deprecated in the future.

        Parameters
        ----------
        success : bool, optional
           if the action succeeded.
        """
        if success:
            self.set_finished()
        else:
            # success=False does not give any information about *why* it
            # failed, so set a generic exception.
            exc = UnknownStatusFailure(
                f"The status {self!r} has failed. To obtain more specific, "
                "helpful errors in the future, update the Device to use "
                "set_exception(...) instead of _finished(success=False)."
            )
            self.set_exception(exc)

    def exception(self, timeout=None):
        """
        Return the exception raised by the action.

        If the action has completed successfully, return ``None``. If it has
        finished in error, return the exception.

        Parameters
        ----------
        timeout: Union[Number, None], optional
            If None (default) wait indefinitely until the status finishes.

        Raises
        ------
        WaitTimeoutError
            If the status has not completed within ``timeout`` (starting from
            when this method was called, not from the beginning of the action).
        """
        if not self._event.wait(timeout=timeout):
            raise WaitTimeoutError("Status has not completed yet.")
        return self._exception

    def wait(self, timeout=None):
        """
        Block until the action completes.

        When the action has finished succesfully, return ``None``. If the
        action has failed, raise the exception.

        Parameters
        ----------
        timeout: Union[Number, None], optional
            If None (default) wait indefinitely until the status finishes.

        Raises
        ------
        WaitTimeoutError
            If the status has not completed within ``timeout`` (starting from
            when this method was called, not from the beginning of the action).
        StatusTimeoutError
            If the status has failed because the *timeout* that it was
            initialized with has expired.
        Exception
            This is ``status.exception()``, raised if the status has finished
            with an error.  This may include ``TimeoutError``, which
            indicates that the action itself raised ``TimeoutError``, distinct
            from ``WaitTimeoutError`` above.
        """
        if not self._event.wait(timeout=timeout):
            raise WaitTimeoutError("Status has not completed yet.")
        if self._exception is not None:
            raise self._exception

    @property
    def callbacks(self):
        """
        Callbacks to be run when the status is marked as finished
        """
        return self._callbacks

    @property
    def finished_cb(self):
        with self._lock:
            if len(self.callbacks) == 1:
                warn(
                    "The property `finished_cb` is deprecated, and must raise "
                    "an error if a status object has multiple callbacks. Use "
                    "the `callbacks` property instead.",
                    stacklevel=2,
                )
                (cb,) = self.callbacks
                assert cb is not None
                return cb
            else:
                raise UseNewProperty(
                    "The deprecated `finished_cb` property "
                    "cannot be used for status objects that have "
                    "multiple callbacks. Use the `callbacks` "
                    "property instead."
                )

    def add_callback(self, callback):
        """
        Register a callback to be called once when the Status finishes.

        The callback will be called exactly once. If the Status is finished
        before a callback is added, it will be called immediately. This is
        threadsafe.

        The callback will be called regardless of success of failure. The
        callback has access to this status object, so it can distinguish success
        or failure by inspecting the object.

        Parameters
        ----------
        callback: callable
            Expected signature: ``callback(status)``.

            The signature ``callback()`` is also supported for
            backward-compatibility but will issue warnings. Support will be
            removed in a future release of ophyd.
        """
        # Handle func with signature callback() for back-compat.
        callback = adapt_old_callback_signature(callback)
        with self._lock:
            if self.done:
                # Call it once and do not hold a reference to it.
                callback(self)
            else:
                # Hold a strong reference to this. In other contexts we tend to
                # hold weak references to callbacks, but this is a single-shot
                # callback, so we will hold a strong reference until we call it,
                # and then clear this cache to drop the reference(s).
                self._callbacks.append(callback)

    @finished_cb.setter
    def finished_cb(self, cb):
        with self._lock:
            if not self.callbacks:
                warn(
                    "The setter `finished_cb` is deprecated, and must raise "
                    "an error if a status object already has one callback. Use "
                    "the `add_callback` method instead.",
                    stacklevel=2,
                )
                self.add_callback(cb)
            else:
                raise UseNewProperty(
                    "The deprecated `finished_cb` setter cannot "
                    "be used for status objects that already "
                    "have one callback. Use the `add_callbacks` "
                    "method instead."
                )

    def __and__(self, other):
        """
        Returns a new 'composite' status object, AndStatus,
        with the same base API.

        It will finish when both `self` or `other` finish.
        """
        return AndStatus(self, other)


class AndStatus(StatusBase):
    "a Status that has composes two other Status objects using logical and"

    def __init__(self, left, right, **kwargs):
        super().__init__(**kwargs)
        self.left = left
        self.right = right

        def inner(status):
            with self._lock:
                if self._externally_initiated_completion:
                    return
                with self.left._lock:
                    with self.right._lock:
                        l_success = self.left.success
                        r_success = self.right.success
                        l_done = self.left.done
                        r_done = self.right.done

                        # At least one is done.
                        # If it failed, do not wait for the second one.
                        if (not l_success) and l_done:
                            self._finished(success=False)
                        elif (not r_success) and r_done:
                            self._finished(success=False)

                        elif l_success and r_success and l_done and r_done:
                            # Both are done, successfully.
                            self._finished(success=True)
                        # Else one is done, successfully, and we wait for #2,
                        # when this function will be called again.

        self.left.add_callback(inner)
        self.right.add_callback(inner)

    def __repr__(self):
        return "({self.left!r} & {self.right!r})".format(self=self)

    def __str__(self):
        return (
            "{0}(done={1.done}, "
            "success={1.success})"
            "".format(self.__class__.__name__, self)
        )


class Status(StatusBase):
    """
    Track the status of a potentially-lengthy action like moving or triggering.

    This has room for an option ``obj`` parameter, noting the object associated
    with action. Status does not use this internally, but it can be useful for
    external code to keep track of things.

    Parameters
    ----------
    timeout: float, optional
        The amount of time to wait before marking the Status as failed.  If
        ``None`` (default) wait forever. It is strongly encouraged to set a
        finite timeout.  If settle_time below is set, that time is added to the
        effective timeout.
    settle_time: float, optional
        The amount of time to wait between the caller specifying that the
        status has completed to running callbacks. Default is 0.

    Attributes
    ----------
    obj : any or None
        The object
    """

    def __init__(self, obj=None, timeout=None, settle_time=0, done=None, success=None):
        self.obj = obj
        super().__init__(
            timeout=timeout, settle_time=settle_time, done=done, success=success
        )

    def __str__(self):
        return (
            "{0}(obj={1.obj}, "
            "done={1.done}, "
            "success={1.success})"
            "".format(self.__class__.__name__, self)
        )

    __repr__ = __str__


class DeviceStatus(StatusBase):
    """
    Track the status of a potentially-lengthy action like moving or triggering.

    This adds the notion of a Device and minimal support for progress bars.
    (They only get notified of the Device name and the time of completion.)
    See MoveStatus for a richer implementation of progress bars.

    Parameters
    ----------
    timeout: float, optional
        The amount of time to wait before marking the Status as failed.  If
        ``None`` (default) wait forever. It is strongly encouraged to set a
        finite timeout.  If settle_time below is set, that time is added to the
        effective timeout.
    settle_time: float, optional
        The amount of time to wait between the caller specifying that the
        status has completed to running callbacks. Default is 0.
    """

    def __init__(self, device, **kwargs):
        self.device = device
        self._watchers = []
        super().__init__(**kwargs)

    def _handle_failure(self):
        super()._handle_failure()
        self.log.debug("Trying to stop %s", repr(self.device))
        self.device.stop()

    def __str__(self):
        return (
            "{0}(device={1.device.name}, done={1.done}, "
            "success={1.success})"
            "".format(self.__class__.__name__, self)
        )

    def watch(self, func):
        """
        Subscribe to notifications about partial progress.
        """
        # See MoveStatus.watch for a richer implementation and more info.
        if self.device is not None:
            self._watchers.append(func)
            func(name=self.device.name)

    def _settled(self):
        """Hook for when status has completed and settled"""
        for watcher in self._watchers:
            watcher(name=self.device.name, fraction=0.0)

    __repr__ = __str__


class SubscriptionStatus(DeviceStatus):
    """
    Status updated via `ophyd` events

    Parameters
    ----------
    device : obj

    callback : callable
        Callback that takes event information and returns a boolean. Signature
        should be ``f(*, old_value, value, **kwargs)``. The arguments
        old_value and value will be passed in by keyword, so their order does
        not matter.

    event_type : str, optional
        Name of event type to check whether the device has finished succesfully

    timeout : float, optional
        Maximum timeout to wait to mark the request as a failure

    settle_time : float, optional
        Time to wait after completion until running callbacks

    run: bool, optional
        Run the callback now
    """

    def __init__(
        self,
        device,
        callback,
        event_type=None,
        timeout=None,
        settle_time=None,
        run=True,
    ):
        # Store device and attribute information
        self.device = device
        self.callback = callback

        # Start timeout thread in the background
        super().__init__(device, timeout=timeout, settle_time=settle_time)

        # Subscribe callback and run initial check
        self.device.subscribe(self.check_value, event_type=event_type, run=run)

    def check_value(self, *args, **kwargs):
        """
        Update the status object
        """
        # Get attribute from device
        try:
            success = self.callback(*args, **kwargs)

        # Do not fail silently
        except Exception as e:
            self.log.error(e)
            raise

        # If successfull indicate completion
        if success:
            self._finished(success=True)

    def set_finished(self):
        """
        Mark as finished successfully.

        This method should generally not be called by the *recipient* of this
        Status object, but only by the object that created and returned it.
        """
        # Clear callback
        self.device.clear_sub(self.check_value)
        # Run completion
        super().set_finished()

    def _handle_failure(self):
        # This is called whether we fail via the timeout thread or via an
        # a call to set_exception.
        # Clear callback
        self.device.clear_sub(self.check_value)
        return super()._handle_failure()


class MoveStatus(DeviceStatus):
    """
    Track the state of a movement from some initial to final "position".

    The position could a physical position, a "position" in a pseudo-space, a
    temperature, etc. This constraint allows richer support for progress bars,
    including progress updates and an ETA.

    Parameters
    ----------
    positioner : Positioner
    target : float or array-like
        Target position
    done : bool, optional
        Whether or not the motion has already completed
    success : bool, optional
        If motion has already completed, the status of that motion
    start_ts : float, optional
        The motion start timestamp
    timeout : float, optional
        The default timeout to use for a blocking wait, and the amount of time
        to wait to mark the motion as failed
    settle_time : float, optional
        The amount of time to wait between motion completion and running
        callbacks

    Attributes
    ----------
    pos : Positioner
    target : float or array-like
        Target position
    done : bool
        Whether or not the motion has already completed
    start_ts : float
        The motion start timestamp
    finish_ts : float
        The motion completd timestamp
    finish_pos : float or ndarray
        The final position
    success : bool
        Motion successfully completed
    """

    def __init__(self, positioner, target, *, start_ts=None, **kwargs):
        self._tname = "timeout for {}".format(positioner.name)
        if start_ts is None:
            start_ts = time.time()

        self.pos = positioner
        self.target = target
        self.start_ts = start_ts
        self.start_pos = self.pos.position
        self.finish_ts = None
        self.finish_pos = None

        self._unit = getattr(self.pos, "egu", None)
        self._precision = getattr(self.pos, "precision", None)
        self._name = self.pos.name

        # call the base class
        super().__init__(positioner, **kwargs)

        # Notify watchers (things like progress bars) of new values
        # at the device's natural update rate.
        if not self.done:
            self.pos.subscribe(self._notify_watchers, event_type=self.pos.SUB_READBACK)

    def watch(self, func):
        """
        Subscribe to notifications about partial progress.

        This is useful for progress bars.

        Parameters
        ----------
        func : callable
            Expected to accept the keyword aruments:

                * ``name``
                * ``current``
                * ``initial``
                * ``target``
                * ``unit``
                * ``precision``
                * ``fraction``
                * ``time_elapsed``
                * ``time_remaining``
        """
        self._watchers.append(func)

    def _notify_watchers(self, value, *args, **kwargs):
        # *args and **kwargs catch extra inputs from pyepics, not needed here
        if not self._watchers:
            return
        current = value
        target = self.target
        initial = self.start_pos
        time_elapsed = time.time() - self.start_ts
        try:
            fraction = np.clip(abs(target - current) / abs(initial - target), 0, 1)
        # maybe we can't do math?
        except (TypeError, ZeroDivisionError):
            fraction = None

        if fraction is not None and np.isnan(fraction):
            fraction = None

        for watcher in self._watchers:
            watcher(
                name=self._name,
                current=current,
                initial=initial,
                target=target,
                unit=self._unit,
                precision=self._precision,
                time_elapsed=time_elapsed,
                fraction=fraction,
            )

    @property
    def error(self):
        """Error between target position and current* position

        * If motion is already complete, the final position is used
        """
        if self.finish_pos is not None:
            finish_pos = self.finish_pos
        else:
            finish_pos = self.pos.position

        try:
            return np.array(finish_pos) - np.array(self.target)
        except Exception:
            return None

    def _settled(self):
        """Hook for when motion has completed and settled"""
        super()._settled()
        self.pos.clear_sub(self._notify_watchers)
        self._watchers.clear()
        self.finish_ts = time.time()
        self.finish_pos = self.pos.position

    @property
    def elapsed(self):
        """Elapsed time"""
        if self.finish_ts is None:
            return time.time() - self.start_ts
        else:
            return self.finish_ts - self.start_ts

    def __str__(self):
        return (
            "{0}(done={1.done}, pos={1.pos.name}, "
            "elapsed={1.elapsed:.1f}, "
            "success={1.success}, settle_time={1.settle_time})"
            "".format(self.__class__.__name__, self)
        )

    __repr__ = __str__


def wait(status, timeout=None, *, poll_rate="DEPRECATED"):
    """(Blocking) wait for the status object to complete

    Parameters
    ----------
    status: StatusBase
        A Status object
    timeout: Union[Number, None], optional
        Amount of time in seconds to wait. None disables, such that wait() will
        only return when either the status completes or if interrupted by the
        user.
    poll_rate: "DEPRECATED"
        DEPRECATED. Has no effect because this does not poll.

    Raises
    ------
    WaitTimeoutError
        If the status has not completed within ``timeout`` (starting from
        when this method was called, not from the beginning of the action).
    Exception
        This is ``status.exception()``, raised if the status has finished
        with an error.  This may include ``TimeoutError``, which
        indicates that the action itself raised ``TimeoutError``, distinct
        from ``WaitTimeoutError`` above.
    """
    return status.wait(timeout)



import threading

import atexit
import logging

from caproto.threading import pyepics_compat
from caproto.threading.pyepics_compat import PV as _PV, caput, caget  # noqa
from ._dispatch import _CallbackThread, EventDispatcher, wrap_callback


thread_class = threading.Thread
module_logger = logging.getLogger(__name__)
_dispatcher = None
name = "caproto"


def get_dispatcher():
    "The event dispatcher for the caproto control layer"
    return _dispatcher


class CaprotoCallbackThread(_CallbackThread):
    ...


class PV(_PV):
    def __init__(
        self,
        pvname,
        callback=None,
        form="time",
        verbose=False,
        auto_monitor=None,
        count=None,
        connection_callback=None,
        connection_timeout=None,
        access_callback=None,
        context=None,
    ):
        connection_callback = wrap_callback(
            _dispatcher, "metadata", connection_callback
        )
        callback = wrap_callback(_dispatcher, "monitor", callback)
        access_callback = wrap_callback(_dispatcher, "metadata", access_callback)

        super().__init__(
            pvname,
            form=form,
            verbose=verbose,
            auto_monitor=auto_monitor,
            count=count,
            connection_timeout=connection_timeout,
            connection_callback=connection_callback,
            callback=callback,
            access_callback=access_callback,
            context=context,
        )

    def add_callback(
        self, callback=None, index=None, run_now=False, with_ctrlvars=True, **kw
    ):
        if not self.auto_monitor:
            self.auto_monitor = True
        callback = wrap_callback(_dispatcher, "monitor", callback)
        return super().add_callback(
            callback=callback,
            index=index,
            run_now=run_now,
            with_ctrlvars=with_ctrlvars,
            **kw
        )

    def put(
        self,
        value,
        wait=False,
        timeout=30.0,
        use_complete=False,
        callback=None,
        callback_data=None,
    ):
        if callback:
            use_complete = True
        callback = wrap_callback(_dispatcher, "get_put", callback)
        return super().put(
            value,
            wait=wait,
            timeout=timeout,
            use_complete=use_complete,
            callback=callback,
            callback_data=callback_data,
        )

    # TODO: caproto breaks API compatibility in wait_for_connection, raising TimeoutError

    def get_all_metadata_blocking(self, timeout):
        if self._args["status"] is None:
            self.get_timevars(timeout=timeout)
        self.get_ctrlvars(timeout=timeout)
        md = self._args.copy()
        md.pop("value", None)
        return md

    def get_all_metadata_callback(self, callback, *, timeout):
        def get_metadata_thread(pvname):
            md = self.get_all_metadata_blocking(timeout=timeout)
            callback(pvname, md)

        _dispatcher.schedule_utility_task(get_metadata_thread, pvname=self.pvname)

    def clear_callbacks(self):
        super().clear_callbacks()
        self.access_callbacks.clear()
        self.connection_callbacks.clear()

    def clear_auto_monitor(self):
        # TODO move into caproto
        self.auto_monitor = False
        if self._auto_monitor_sub is not None:
            self._auto_monitor_sub.clear()
            self._auto_monitor_sub = None


def release_pvs(*pvs):
    for pv in pvs:
        pv.clear_callbacks()
        # pv.disconnect()


def get_pv(
    pvname,
    form="time",
    connect=False,
    context=None,
    timeout=5.0,
    connection_callback=None,
    access_callback=None,
    callback=None,
    **kwargs
):
    """Get a PV from PV cache or create one if needed.

    Parameters
    ---------
    form : str, optional
        PV form: one of 'native' (default), 'time', 'ctrl'
    connect : bool, optional
        whether to wait for connection (default False)
    context : int, optional
        PV threading context (defaults to current context)
    timeout : float, optional
        connection timeout, in seconds (default 5.0)
    """
    if context is None:
        context = PV._default_context

    pv = PV(
        pvname,
        form=form,
        connection_callback=connection_callback,
        access_callback=access_callback,
        callback=callback,
        **kwargs
    )
    pv._reference_count = 0
    if connect:
        pv.wait_for_connection(timeout=timeout)
    return pv


def setup(logger):
    """Setup ophyd for use

    Must be called once per session using ophyd
    """
    # It's important to use the same context in the callback _dispatcher
    # as the main thread, otherwise not-so-savvy users will be very
    # confused
    global _dispatcher

    if _dispatcher is not None:
        logger.debug("ophyd already setup")
        return

    pyepics_compat._get_pv = pyepics_compat.get_pv
    pyepics_compat.get_pv = get_pv

    def _cleanup():
        """Clean up the ophyd session"""
        global _dispatcher
        if _dispatcher is None:
            return

        pyepics_compat.get_pv = pyepics_compat._get_pv

        if _dispatcher.is_alive():
            _dispatcher.stop()

        _dispatcher = None

    logger.debug("Installing event dispatcher")
    context = PV._default_context.broadcaster
    _dispatcher = EventDispatcher(
        thread_class=CaprotoCallbackThread, context=context, logger=logger
    )
    atexit.register(_cleanup)
    return _dispatcher



import threading
import unittest.mock

import pytest

import ophyd.sim
import ophyd.units


class MockCallbackHelper:
    """
    Simple helper for getting a callback, setting an event, and checking args.

    Use __call__ as the hook and inspect/verify ``mock`` or ``call_kwargs``.
    """

    def __init__(self):
        self.event = threading.Event()
        self.mock = unittest.mock.Mock()

    def __call__(self, *args, **kwargs):
        self.mock(*args, **kwargs)
        self.event.set()

    def wait(self, timeout=1.0):
        """Wait for the callback to be called."""
        self.event.wait(timeout)

    @property
    def call_kwargs(self):
        """Call keyword arguments."""
        _, kwargs = self.mock.call_args
        return kwargs


@pytest.fixture(scope="function")
def unit_conv_signal():
    orig = ophyd.sim.FakeEpicsSignal("sig", name="orig")
    assert "units" in orig.metadata_keys

    orig.sim_put(5)

    return ophyd.units.UnitConversionDerivedSignal(
        derived_from=orig,
        original_units="m",
        derived_units="mm",
        name="converted",
    )


def test_unit_conversion_signal_units(unit_conv_signal):
    assert unit_conv_signal.original_units == "m"
    assert unit_conv_signal.derived_units == "mm"
    assert unit_conv_signal.describe()[unit_conv_signal.name]["units"] == "mm"


def test_unit_conversion_signal_get_put(unit_conv_signal):
    assert unit_conv_signal.get() == 5_000
    unit_conv_signal.put(10_000, wait=True)
    assert unit_conv_signal.derived_from.get() == 10


def test_unit_conversion_signal_value_sub(unit_conv_signal):
    helper = MockCallbackHelper()
    unit_conv_signal.subscribe(helper, run=False)
    unit_conv_signal.derived_from.put(20, wait=True)
    helper.wait(1)
    helper.mock.assert_called_once()

    assert helper.call_kwargs["value"] == 20_000
    assert unit_conv_signal.get() == 20_000


def test_unit_conversion_signal_metadata_sub(unit_conv_signal):
    helper = MockCallbackHelper()
    unit_conv_signal.subscribe(helper, run=True, event_type="meta")
    helper.wait(1)
    helper.mock.assert_called_once()
    assert helper.call_kwargs["units"] == "mm"



import time
from unittest.mock import Mock

from ophyd import Device
from ophyd.status import StatusBase, SubscriptionStatus, UseNewProperty, MoveStatus
from ophyd.utils import (
    InvalidState,
    UnknownStatusFailure,
    StatusTimeoutError,
    WaitTimeoutError,
)
import pytest


def _setup_state_and_cb(new_signature=True):
    state = {}

    if new_signature:

        def cb(status):
            state["done"] = True

    else:

        def cb():
            state["done"] = True

    return state, cb


def test_status_post():
    st = StatusBase()
    state, cb = _setup_state_and_cb()

    assert "done" not in state
    st.add_callback(cb)
    assert "done" not in state
    st.set_finished()
    st.wait(1)
    time.sleep(0.1)  # Wait for callbacks to run.
    assert "done" in state
    assert state["done"]


def test_status_legacy_finished_cb():
    st = StatusBase()
    state1, cb1 = _setup_state_and_cb()
    state2, cb2 = _setup_state_and_cb()

    # The old setter works for adding one callback.
    with pytest.warns(UserWarning):
        st.finished_cb = cb1
    # The new getter works.
    st.callbacks == set([cb1])
    # And the old getter works so long as there is just one callback set.
    with pytest.warns(UserWarning):
        assert st.finished_cb is cb1
    # As before, the old setter cannot be updated once set.
    with pytest.raises(UseNewProperty):
        st.finished_cb = cb2
    # But, using the new method, we can add another callback.
    st.add_callback(cb2)
    # Once we have two callbacks, the getter does not work.
    with pytest.raises(UseNewProperty):
        st.finished_cb
    # But the new getter does.
    st.callbacks == set([cb1, cb2])

    assert "done" not in state1
    assert "done" not in state2
    st.set_finished()
    st.wait(1)
    time.sleep(0.1)  # Wait for callbacks to run.
    assert "done" in state1
    assert "done" in state2


def test_status_pre():
    st = StatusBase()
    state, cb = _setup_state_and_cb()

    st.set_finished()
    st.wait(1)
    time.sleep(0.1)  # Wait for callbacks to run.

    assert "done" not in state
    st.add_callback(cb)
    assert "done" in state
    assert state["done"]


def test_direct_done_setting():
    st = StatusBase()
    state, cb = _setup_state_and_cb()

    with pytest.raises(RuntimeError):
        st.done = True  # changing isn't allowed
    with pytest.warns(UserWarning):
        st.done = False  # but for now no-ops warn

    st.set_finished()
    st.wait(1)
    time.sleep(0.1)  # Wait for callbacks to run.

    with pytest.raises(RuntimeError):
        st.done = False  # changing isn't allowed
    with pytest.warns(UserWarning):
        st.done = True  # but for now no-ops warn


def test_subscription_status():
    # Arbitrary device
    d = Device("Tst:Prefix", name="test")
    # Mock callback
    m = Mock()

    # Full fake callback signature
    def cb(*args, done=False, **kwargs):
        # Run mock callback
        m()
        # Return finished or not
        return done

    status = SubscriptionStatus(d, cb, event_type=d.SUB_ACQ_DONE)

    # Run callbacks but do not mark as complete
    d._run_subs(sub_type=d.SUB_ACQ_DONE, done=False)
    time.sleep(0.1)  # Wait for callbacks to run.
    assert m.called
    assert not status.done and not status.success

    # Run callbacks and mark as complete
    d._run_subs(sub_type=d.SUB_ACQ_DONE, done=True)
    time.sleep(0.1)  # Wait for callbacks to run.
    assert status.done and status.success


def test_and():
    st1 = StatusBase()
    st2 = StatusBase()
    st3 = st1 & st2
    # make sure deep recursion works
    st4 = st1 & st3
    st5 = st3 & st4
    state1, cb1 = _setup_state_and_cb()
    state2, cb2 = _setup_state_and_cb()
    state3, cb3 = _setup_state_and_cb()
    state4, cb4 = _setup_state_and_cb()
    state5, cb5 = _setup_state_and_cb()
    st1.add_callback(cb1)
    st2.add_callback(cb2)
    st3.add_callback(cb3)
    st4.add_callback(cb4)
    st5.add_callback(cb5)
    st1.set_finished()
    st1.wait(1)
    time.sleep(0.1)  # Wait for callbacks to run.
    assert "done" in state1
    assert "done" not in state2
    assert "done" not in state3
    assert "done" not in state4
    assert "done" not in state5
    st2.set_finished()
    st3.wait(1)
    st4.wait(1)
    st5.wait(1)
    time.sleep(0.1)  # Wait for callbacks to run.
    assert "done" in state3
    assert "done" in state4
    assert "done" in state5
    assert st3.left is st1
    assert st3.right is st2
    assert st4.left is st1
    assert st4.right is st3
    assert st5.left is st3
    assert st5.right is st4


def test_notify_watchers():
    from ophyd.sim import hw

    hw = hw()
    mst = MoveStatus(hw.motor, 10)

    def callback(*args, **kwargs):
        ...

    mst.watch(callback)
    mst.target = 0
    mst.start_pos = 0
    mst._notify_watchers(0)


def test_old_signature():
    st = StatusBase()
    state, cb = _setup_state_and_cb(new_signature=False)
    with pytest.warns(DeprecationWarning, match="signature"):
        st.add_callback(cb)
    assert not state
    st.set_finished()
    st.wait(1)
    time.sleep(0.1)  # Wait for callbacks to run.
    assert state


def test_old_signature_on_finished_status():
    st = StatusBase()
    state, cb = _setup_state_and_cb(new_signature=False)
    st.set_finished()
    st.wait(1)
    with pytest.warns(DeprecationWarning, match="signature"):
        st.add_callback(cb)
    assert state


def test_old_finished_method_success():
    st = StatusBase()
    state, cb = _setup_state_and_cb()
    st.add_callback(cb)
    assert not state
    st._finished()
    st.wait(1)
    time.sleep(0.1)  # Wait for callbacks to run.
    assert state
    assert st.done
    assert st.success


def test_old_finished_method_failure():
    st = StatusBase()
    state, cb = _setup_state_and_cb()
    st.add_callback(cb)
    assert not state
    st._finished(success=False)
    with pytest.raises(UnknownStatusFailure):
        st.wait(1)
    time.sleep(0.1)  # Wait for callbacks to run.
    assert state
    assert st.done
    assert not st.success


def test_set_finished_twice():
    st = StatusBase()
    st.set_finished()
    with pytest.raises(InvalidState):
        st.set_finished()


def test_set_exception_twice():
    st = StatusBase()
    exc = Exception()
    st.set_exception(exc)
    with pytest.raises(InvalidState):
        st.set_exception(exc)


def test_set_exception_wrong_type():
    st = StatusBase()
    NOT_AN_EXCEPTION = object()
    with pytest.raises(ValueError):
        st.set_exception(NOT_AN_EXCEPTION)


def test_set_exception_special_banned_exceptions():
    """
    Exceptions with special significant to StatusBase are banned. See comments
    in set_exception.
    """
    st = StatusBase()
    # Test the class and the instance of each.
    with pytest.raises(ValueError):
        st.set_exception(StatusTimeoutError)
    with pytest.raises(ValueError):
        st.set_exception(StatusTimeoutError())
    with pytest.raises(ValueError):
        st.set_exception(WaitTimeoutError)
    with pytest.raises(ValueError):
        st.set_exception(WaitTimeoutError())


def test_exception_fail_path():
    st = StatusBase()

    class LocalException(Exception):
        ...

    exc = LocalException()
    st.set_exception(exc)
    assert exc is st.exception()
    with pytest.raises(LocalException):
        st.wait(1)


def test_exception_fail_path_with_class():
    """
    Python allows `raise Exception` and `raise Exception()` so we do as well.
    """
    st = StatusBase()

    class LocalException(Exception):
        ...

    st.set_exception(LocalException)
    assert LocalException is st.exception()
    with pytest.raises(LocalException):
        st.wait(1)


def test_exception_success_path():
    st = StatusBase()
    st.set_finished()
    assert st.wait(1) is None
    assert st.exception() is None


def test_wait_timeout():
    """
    A WaitTimeoutError is raised when we block on wait(TIMEOUT) or
    exception(TIMEOUT) and the Status has not finished.
    """
    st = StatusBase()
    with pytest.raises(WaitTimeoutError):
        st.wait(0.01)
    with pytest.raises(WaitTimeoutError):
        st.exception(0.01)


def test_status_timeout():
    """
    A StatusTimeoutError is raised when the timeout set in __init__ has
    expired.
    """
    st = StatusBase(timeout=0)
    with pytest.raises(StatusTimeoutError):
        st.wait(1)
    assert isinstance(st.exception(), StatusTimeoutError)


def test_status_timeout_with_settle_time():
    """
    A StatusTimeoutError is raised when the timeout set in __init__ plus the
    settle_time has expired.
    """
    st = StatusBase(timeout=0, settle_time=1)
    # Not dead yet.
    with pytest.raises(WaitTimeoutError):
        st.exception(0.01)
    # But now we are.
    with pytest.raises(StatusTimeoutError):
        st.wait(2)


def test_external_timeout():
    """
    A TimeoutError is raised, not StatusTimeoutError or WaitTimeoutError,
    when set_exception(TimeoutError) has been set.
    """
    st = StatusBase(timeout=1)
    st.set_exception(TimeoutError())
    with pytest.raises(TimeoutError) as exc:
        st.wait(1)
    assert not isinstance(exc, WaitTimeoutError)
    assert not isinstance(exc, StatusTimeoutError)


def test_race_settle_time_and_timeout():
    """
    A StatusTimeoutError should NOT occur here because that is only invoked
    after (timeout + settle_time) has elapsed.
    """
    st = StatusBase(timeout=1, settle_time=3)
    st.set_finished()  # starts a threading.Timer with the settle_time
    time.sleep(1.5)
    # We should still be settling....
    with pytest.raises(WaitTimeoutError):
        st.wait(1)
    # Now we should be done successfully.
    st.wait(3)


def test_set_finished_after_timeout():
    """
    If an external callback (e.g. pyepics) calls set_finished after the status
    has timed out, ignore it.
    """
    st = StatusBase(timeout=0)
    time.sleep(0.1)
    assert isinstance(st.exception(), StatusTimeoutError)
    # External callback fires, too late.
    st.set_finished()
    assert isinstance(st.exception(), StatusTimeoutError)


def test_set_exception_after_timeout():
    """
    If an external callback (e.g. pyepics) calls set_exception after the status
    has timed out, ignore it.
    """
    st = StatusBase(timeout=0)
    time.sleep(0.1)
    assert isinstance(st.exception(), StatusTimeoutError)

    class LocalException(Exception):
        ...

    # External callback reports failure, too late.
    st.set_exception(LocalException())
    assert isinstance(st.exception(), StatusTimeoutError)


def test_nonsensical_init():
    with pytest.raises(ValueError):
        StatusBase(success=True, done=False)
    with pytest.raises(ValueError):
        StatusBase(success=True, done=None)


def test_deprecated_init():
    with pytest.warns(DeprecationWarning, match="set_finished"):
        StatusBase(success=True, done=True)
    with pytest.warns(DeprecationWarning, match="set_finished"):
        StatusBase(success=False, done=True)
    with pytest.warns(DeprecationWarning, match="set_finished"):
        StatusBase(success=False, done=False)
    with pytest.warns(DeprecationWarning, match="set_finished"):
        StatusBase(success=False, done=None)
    with pytest.warns(DeprecationWarning, match="set_finished"):
        StatusBase(success=None, done=True)
    with pytest.warns(DeprecationWarning, match="set_finished"):
        StatusBase(success=None, done=False)


def test_error_in_settled_method():
    state, cb = _setup_state_and_cb()

    class BrokenStatus(StatusBase):
        def _settled(self):
            raise Exception

    st = BrokenStatus()
    st.add_callback(cb)
    st.set_finished()
    st.wait(1)
    time.sleep(0.1)  # Wait for callbacks to run.
    assert state


def test_error_in_handle_failure_method():
    state, cb = _setup_state_and_cb()

    class BrokenStatus(StatusBase):
        def _handle_failure(self):
            raise Exception

    st = BrokenStatus()
    st.add_callback(cb)
    st.set_finished()
    st.wait(1)
    time.sleep(0.1)  # Wait for callbacks to run.
    assert state



""" """

import logging

logger = logging.getLogger(__name__)

from .base import *  # noqa: F401, F402, E402, F403
from .cam import *  # noqa: F401, F402, E402, F403
from .detectors import *  # noqa: F401, F402, E402, F403

# NOTE: the following imports are here for backward compatibility with
# previous ophyd versions. This does not represent all available plugins
# in ophyd. For that, import directly from ophyd.areadetector.plugins.
from .plugins import (  # noqa: F401, F402, E402
    ColorConvPlugin,
    FilePlugin,
    HDF5Plugin,
    ImagePlugin,
    JPEGPlugin,
    MagickPlugin,
    NetCDFPlugin,
    NexusPlugin,
    OverlayPlugin,
    ProcessPlugin,
    ROIPlugin,
    StatsPlugin,
    TIFFPlugin,
    TransformPlugin,
    get_areadetector_plugin,
    plugin_from_pvname,
    register_plugin,
)

from .common_plugins import *  # noqa: F401, F402, E402, F403
from .trigger_mixins import *  # noqa: F401, F402, E402, F403
from .paths import EpicsPathSignal  # noqa: F401, F402, E402, F403



import logging
import types
import os
from ophyd.log import set_handler  # noqa: F401

logger = logging.getLogger(__name__)

cl = None


def set_cl(control_layer=None, *, pv_telemetry=False):
    global cl
    known_layers = ("pyepics", "caproto", "dummy")

    if control_layer is None:
        control_layer = os.environ.get("OPHYD_CONTROL_LAYER", "any").lower()

    if control_layer == "any":
        for c_type in known_layers:
            try:
                set_cl(c_type, pv_telemetry=pv_telemetry)
            except ImportError:
                continue
            else:
                return
        else:
            raise ImportError("no valid control layer found")

    # TODO replace this with fancier meta-programming
    # TODO handle control_layer being a module/nampspace directly
    if control_layer == "pyepics":
        from . import _pyepics_shim as shim
    elif control_layer == "caproto":
        from . import _caproto_shim as shim
    elif control_layer == "dummy":
        from . import _dummy_shim as shim
    else:
        raise ValueError("unknown control_layer")

    shim.setup(logger)

    exports = (
        "setup",
        "caput",
        "caget",
        "get_pv",
        "thread_class",
        "name",
        "release_pvs",
        "get_dispatcher",
    )
    # this sets the module level value
    cl = types.SimpleNamespace(**{k: getattr(shim, k) for k in exports})
    if pv_telemetry:
        from functools import wraps
        from collections import Counter

        def decorate_get_pv(func):
            c = Counter()

            @wraps(func)
            def get_pv(pvname, *args, **kwargs):
                c[pvname] += 1
                return func(pvname, *args, **kwargs)

            get_pv.counter = c
            return get_pv

        cl.get_pv = decorate_get_pv(cl.get_pv)


def get_cl():
    if cl is None:
        raise RuntimeError("control layer not set, " "unsure how you got to this state")
    return cl


set_cl()

from .ophydobj import (  # noqa: F401, F402, E402
    Kind,
    select_version,
    register_instances_in_weakset,
    register_instances_keyed_on_name,
)

# Signals
from .signal import (  # noqa: F401, F402, E402
    Signal,
    SignalRO,
    EpicsSignal,
    EpicsSignalRO,
    EpicsSignalNoValidation,
    DerivedSignal,
)

# Positioners
from .positioner import PositionerBase, SoftPositioner  # noqa: F401, F402, E402
from .epics_motor import EpicsMotor, MotorBundle  # noqa: F401, F402, E402
from .pv_positioner import (  # noqa: F401, F402, E402
    PVPositioner,
    PVPositionerPC,
    PVPositionerIsClose,
    PVPositionerDone,
)
from .pseudopos import PseudoPositioner, PseudoSingle  # noqa: F401, F402, E402

# Devices
from .scaler import EpicsScaler  # noqa: F401, F402, E402
from .device import (  # noqa: F401, F402, E402
    Device,
    Component,
    FormattedComponent,
    DynamicDeviceComponent,
    ALL_COMPONENTS,
    kind_context,
    wait_for_lazy_connection,
    do_not_wait_for_lazy_connection,
)
from .status import StatusBase, wait  # noqa: F401, F402, E402
from .mca import EpicsMCA, EpicsDXP  # noqa: F401, F402, E402
from .quadem import QuadEM, NSLS_EM, TetrAMM, APS_EM  # noqa: F401, F402, E402

# Areadetector-related
from .areadetector import *  # noqa: F401, F402, E402, F403
from ._version import get_versions  # noqa: F402, E402

from .utils.startup import setup as setup_ophyd  # noqa: F401, F402, E402


__version__ = get_versions()["version"]
del get_versions



"""
Documentation helpers that can be used in packages relying on ophyd devices.
"""

from . import templates
from .utils import (
    autodoc_default_options,
    autosummary_context,
    get_device_info,
    intersphinx_mapping,
    setup,
)

__all__ = [
    "autodoc_default_options",
    "autosummary_context",
    "get_device_info",
    "intersphinx_mapping",
    "setup",
    "templates",
]



import numbers
from typing import Any, Optional, Tuple

import pint

from .signal import DerivedSignal

_global_unit_registry = None


def convert_unit(value: float, from_units: str, to_units: str):
    """
    Convert ``value`` from ``from_units`` to ``to_units`` using ``pint``.

    Parameters
    ----------
    value : float
        The starting value for the conversion.

    from_units : str
        The starting unit of the provided value.

    to_units : str
        The desired unit to convert the value to.

    Returns
    -------
    new_value : float
        The starting value, but converted to the new unit.
    """

    global _global_unit_registry
    if _global_unit_registry is None:
        # NOTE: Instantiating the unit registry is a heavy operation and should
        # not be performed repeatedly.
        _global_unit_registry = pint.UnitRegistry()

    expr = _global_unit_registry.parse_expression(from_units)
    return (value * expr).to(to_units).magnitude


class UnitConversionDerivedSignal(DerivedSignal):
    """
    A DerivedSignal which performs unit conversion.

    Custom units may be specified for the original signal, or if specified, the
    original signal's units may be retrieved upon first connection.

    Parameters
    ----------
    derived_from : Signal or str
        The signal from which this one is derived.  This may be a string
        attribute name that indicates a sibling to use.  When used in a
        ``Device``, this is then simply the attribute name of another
        ``Component``.

    derived_units : str
        The desired units to use for this signal.  These can also be referred
        to as the "user-facing" units.

    original_units : str, optional
        The units from the original signal.  If not specified, control system
        information regarding units will be retrieved upon first connection.

    user_offset : any, optional
        An optional user offset that will be *subtracted* when updating the
        original signal, and *added* when calculating the derived value.
        This offset should be supplied in ``derived_units`` and not
        ``original_units``.

        For example, if the original signal updates to a converted value of
        500 ``derived_units`` and the ``user_offset`` is set to 100, this
        ``DerivedSignal`` will show a value of 600.  When providing a new
        setpoint, the ``user_offset`` will be subtracted.

    write_access : bool, optional
        Write access may be disabled by setting this to ``False``, regardless
        of the write access of the underlying signal.

    name : str, optional
        The signal name.

    parent : Device, optional
        The parent device.  Required if ``derived_from`` is an attribute name.

    limits : 2-tuple, optional
        Ophyd signal-level limits in derived units.  DerivedSignal defaults
        to converting the original signal's limits, but these may be overridden
        here without modifying the original signal.

    **kwargs :
        Keyword arguments are passed to the superclass.
    """

    derived_units: str
    original_units: str

    def __init__(
        self,
        derived_from,
        *,
        derived_units: str,
        original_units: Optional[str] = None,
        user_offset: Optional[numbers.Real] = 0,
        limits: Optional[Tuple[numbers.Real, numbers.Real]] = None,
        **kwargs,
    ):
        self.derived_units = derived_units
        self.original_units = original_units
        self._user_offset = user_offset
        self._custom_limits = limits
        super().__init__(derived_from, **kwargs)
        self._metadata["units"] = derived_units

        # Ensure that we include units in metadata callbacks, even if the
        # original signal does not include them.
        if "units" not in self._metadata_keys:
            self._metadata_keys = self._metadata_keys + ("units",)

    def forward(self, value):
        """Compute derived signal value -> original signal value"""
        if self.user_offset is None:
            raise ValueError(
                f"{self.name}.user_offset must be set to a non-None value."
            )
        return convert_unit(
            value - self.user_offset, self.derived_units, self.original_units
        )

    def inverse(self, value):
        """Compute original signal value -> derived signal value"""
        if self.user_offset is None:
            raise ValueError(
                f"{self.name}.user_offset must be set to a non-None value."
            )
        return (
            convert_unit(value, self.original_units, self.derived_units)
            + self.user_offset
        )

    @property
    def limits(self):
        """
        Defaults to limits from the original signal (low, high).

        Limit values may be reversed such that ``low <= value <= high`` after
        performing the calculation.

        Limits may also be overridden here without affecting the original
        signal.
        """
        if self._custom_limits is not None:
            return self._custom_limits

        # Fall back to the superclass derived_from limits:
        return tuple(sorted(self.inverse(v) for v in self._derived_from.limits))

    @limits.setter
    def limits(self, value):
        if value is None:
            self._custom_limits = None
            return

        if len(value) != 2 or value[0] >= value[1]:
            raise ValueError("Custom limits must be a 2-tuple (low, high)")

        self._custom_limits = tuple(value)

    @property
    def user_offset(self) -> Optional[Any]:
        """A user-specified offset in *derived*, user-facing units."""
        return self._user_offset

    @user_offset.setter
    def user_offset(self, offset):
        offset_change = -self._user_offset + offset
        self._user_offset = offset
        self._recalculate_position()
        if self._custom_limits is not None:
            self._custom_limits = (
                self._custom_limits[0] + offset_change,
                self._custom_limits[1] + offset_change,
            )

    def _recalculate_position(self):
        """
        Recalculate the derived position and send subscription updates.

        No-operation if the original signal is not connected.
        """
        if not self._derived_from.connected:
            return

        value = self._derived_from.get()
        if value is not None:
            # Note: no kwargs here; no metadata updates
            self._derived_value_callback(value)

    def _derived_metadata_callback(self, *, connected, **kwargs):
        if connected and "units" in kwargs:
            if self.original_units is None:
                self.original_units = kwargs["units"]
        # Do not pass through units, as we have our own.
        kwargs["units"] = self.derived_units
        super()._derived_metadata_callback(connected=connected, **kwargs)

    def describe(self):
        full_desc = super().describe()
        desc = full_desc[self.name]
        desc["units"] = self.derived_units
        # Note: this should be handled in ophyd:
        for key in ("lower_ctrl_limit", "upper_ctrl_limit"):
            if key in desc:
                desc[key] = self.inverse(desc[key])
        return full_desc



import logging

from .utils.epics_pvs import fmt_time

from .signal import EpicsSignal, EpicsSignalRO
from .utils import DisconnectedError
from .utils.epics_pvs import raise_if_disconnected, AlarmSeverity
from .positioner import PositionerBase
from .device import Device, Component as Cpt, required_for_connection
from .status import wait as status_wait
from enum import Enum


logger = logging.getLogger(__name__)


class HomeEnum(str, Enum):
    forward = "forward"
    reverse = "reverse"


class EpicsMotor(Device, PositionerBase):
    """An EPICS motor record, wrapped in a :class:`Positioner`

    Keyword arguments are passed through to the base class, Positioner

    Parameters
    ----------
    prefix : str
        The record to use
    read_attrs : sequence of attribute names
        The signals to be read during data acquisition (i.e., in read() and
        describe() calls)
    name : str, optional
        The name of the device
    parent : instance or None
        The instance of the parent device, if applicable
    settle_time : float, optional
        The amount of time to wait after moves to report status completion
    timeout : float, optional
        The default timeout to use for motion requests, in seconds.
    """

    # position
    user_readback = Cpt(EpicsSignalRO, ".RBV", kind="hinted", auto_monitor=True)
    user_setpoint = Cpt(EpicsSignal, ".VAL", limits=True, auto_monitor=True)

    # calibration dial <-> user
    user_offset = Cpt(EpicsSignal, ".OFF", kind="config", auto_monitor=True)
    user_offset_dir = Cpt(EpicsSignal, ".DIR", kind="config", auto_monitor=True)
    offset_freeze_switch = Cpt(EpicsSignal, ".FOFF", kind="omitted", auto_monitor=True)
    set_use_switch = Cpt(EpicsSignal, ".SET", kind="omitted", auto_monitor=True)

    # configuration
    velocity = Cpt(EpicsSignal, ".VELO", kind="config", auto_monitor=True)
    acceleration = Cpt(EpicsSignal, ".ACCL", kind="config", auto_monitor=True)
    motor_egu = Cpt(EpicsSignal, ".EGU", kind="config", auto_monitor=True)

    # motor status
    motor_is_moving = Cpt(EpicsSignalRO, ".MOVN", kind="omitted", auto_monitor=True)
    motor_done_move = Cpt(EpicsSignalRO, ".DMOV", kind="omitted", auto_monitor=True)
    high_limit_switch = Cpt(EpicsSignalRO, ".HLS", kind="omitted", auto_monitor=True)
    low_limit_switch = Cpt(EpicsSignalRO, ".LLS", kind="omitted", auto_monitor=True)
    high_limit_travel = Cpt(EpicsSignal, ".HLM", kind="omitted", auto_monitor=True)
    low_limit_travel = Cpt(EpicsSignal, ".LLM", kind="omitted", auto_monitor=True)
    direction_of_travel = Cpt(EpicsSignal, ".TDIR", kind="omitted", auto_monitor=True)

    # commands
    motor_stop = Cpt(EpicsSignal, ".STOP", kind="omitted")
    home_forward = Cpt(EpicsSignal, ".HOMF", kind="omitted")
    home_reverse = Cpt(EpicsSignal, ".HOMR", kind="omitted")

    # alarm information
    tolerated_alarm = AlarmSeverity.NO_ALARM

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        **kwargs
    ):
        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs
        )

        # Make the default alias for the user_readback the name of the
        # motor itself.
        self.user_readback.name = self.name

        def on_limit_changed(value, old_value, **kwargs):
            """
            update EpicsSignal object when a limit CA monitor received from EPICS
            """
            if self.connected and old_value is not None and value != old_value:
                self.user_setpoint._metadata_changed(
                    self.user_setpoint.pvname,
                    self.user_setpoint._read_pv.get_ctrlvars(),
                    from_monitor=True,
                    update=True,
                )

        self.low_limit_travel.subscribe(on_limit_changed)
        self.high_limit_travel.subscribe(on_limit_changed)

    @property
    def precision(self):
        """The precision of the readback PV, as reported by EPICS"""
        return self.user_readback.precision

    @property
    def egu(self):
        """The engineering units (EGU) for a position"""
        return self.motor_egu.get()

    @property
    @raise_if_disconnected
    def limits(self):
        return self.user_setpoint.limits

    @property
    @raise_if_disconnected
    def moving(self):
        """Whether or not the motor is moving

        Returns
        -------
        moving : bool
        """
        return bool(self.motor_is_moving.get(use_monitor=False))

    @raise_if_disconnected
    def stop(self, *, success=False):
        self.motor_stop.put(1, wait=False)
        super().stop(success=success)

    @raise_if_disconnected
    def move(self, position, wait=True, **kwargs):
        """Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to
        moved_cb : callable
            Call this callback when movement has finished. This callback must
            accept one keyword argument: 'obj' which will be set to this
            positioner instance.
        timeout : float, optional
            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        Returns
        -------
        status : MoveStatus

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        """
        self._started_moving = False

        status = super().move(position, **kwargs)
        self.user_setpoint.put(position, wait=False)
        try:
            if wait:
                status_wait(status)
        except KeyboardInterrupt:
            self.stop()
            raise

        return status

    @property
    @raise_if_disconnected
    def position(self):
        """The current position of the motor in its engineering units

        Returns
        -------
        position : float
        """
        return self._position

    @raise_if_disconnected
    def set_current_position(self, pos):
        """Configure the motor user position to the given value

        Parameters
        ----------
        pos
           Position to set.

        """
        self.set_use_switch.put(1, wait=True)
        try:
            self.user_setpoint.put(pos, wait=True, force=True)
        finally:
            self.set_use_switch.put(0, wait=True)

    @raise_if_disconnected
    def home(self, direction, wait=True, **kwargs):
        """Perform the default homing function in the desired direction

        Parameters
        ----------
        direction : HomeEnum
           Direction in which to perform the home search.
        """
        direction = HomeEnum(direction)

        self._started_moving = False
        position = (self.low_limit + self.high_limit) / 2
        status = super().move(position, **kwargs)

        if direction == HomeEnum.forward:
            self.home_forward.put(1, wait=False)
        else:
            self.home_reverse.put(1, wait=False)

        try:
            if wait:
                status_wait(status)
        except KeyboardInterrupt:
            self.stop()
            raise

        return status

    def check_value(self, pos):
        """Check that the position is within the soft limits"""
        self.user_setpoint.check_value(pos)

    @required_for_connection
    @user_readback.sub_value
    def _pos_changed(self, timestamp=None, value=None, **kwargs):
        """Callback from EPICS, indicating a change in position"""
        self._set_position(value)

    @required_for_connection
    @motor_done_move.sub_value
    def _move_changed(self, timestamp=None, value=None, sub_type=None, **kwargs):
        """Callback from EPICS, indicating that movement status has changed"""
        was_moving = self._moving
        self._moving = value != 1

        started = False
        if not self._started_moving:
            started = self._started_moving = not was_moving and self._moving

        self.log.debug(
            "[ts=%s] %s moving: %s (value=%s)",
            fmt_time(timestamp),
            self,
            self._moving,
            value,
        )

        if started:
            self._run_subs(
                sub_type=self.SUB_START, timestamp=timestamp, value=value, **kwargs
            )

        if was_moving and not self._moving:
            success = True
            # Check if we are moving towards the low limit switch
            if self.direction_of_travel.get() == 0:
                if self.low_limit_switch.get(use_monitor=False) == 1:
                    success = False
            # No, we are going to the high limit switch
            else:
                if self.high_limit_switch.get(use_monitor=False) == 1:
                    success = False

            # Check the severity of the alarm field after motion is complete.
            # If there is any alarm at all warn the user, and if the alarm is
            # greater than what is tolerated, mark the move as unsuccessful
            severity = self.user_readback.alarm_severity

            if severity != AlarmSeverity.NO_ALARM:
                status = self.user_readback.alarm_status
                if severity > self.tolerated_alarm:
                    self.log.error(
                        "Motion failed: %s is in an alarm state "
                        "status=%s severity=%s",
                        self.name,
                        status,
                        severity,
                    )
                    success = False
                else:
                    self.log.warning(
                        "Motor %s raised an alarm during motion "
                        "status=%s severity %s",
                        self.name,
                        status,
                        severity,
                    )
            self._done_moving(success=success, timestamp=timestamp, value=value)

    @property
    def report(self):
        try:
            rep = super().report
        except DisconnectedError:
            # TODO there might be more in this that gets lost
            rep = {"position": "disconnected"}
        rep["pv"] = self.user_readback.pvname
        return rep

    def get_lim(self, flag):
        """
        Returns the travel limit of motor

        * flag > 0: returns high limit
        * flag < 0: returns low limit
        * flag == 0: returns None

        Included here for compatibility with similar with SPEC command.

        Parameters
        ----------
        high : float
           Limit of travel in the positive direction.
        low : float
           Limit of travel in the negative direction.
        """
        if flag > 0:
            return self.high_limit_travel.get()
        elif flag < 0:
            return self.low_limit_travel.get()

    def set_lim(self, low, high):
        """
        Sets the low and high travel limits of motor

        * No action taken if motor is moving.
        * Low limit is set to lesser of (low, high)
        * High limit is set to greater of (low, high)

        Included here for compatibility with similar with SPEC command.

        Parameters
        ----------
        high : float
           Limit of travel in the positive direction.
        low : float
           Limit of travel in the negative direction.
        """
        if not self.moving:
            # update EPICS
            lo = min(low, high)
            hi = max(low, high)
            if lo <= self.position <= hi:
                self.high_limit_travel.put(hi)
                self.low_limit_travel.put(lo)
                # and ophyd metadata dictionary will update via CA monitor
            else:
                self.log.debug(
                    "Could not set motor limits to (%f, %f) at position %g",
                    low,
                    high,
                    self.position,
                )


class MotorBundle(Device):
    """Sub-class this to device a bundle of motors

    This provides better default behavior for :ref:``hints``.
    """

    ...



from ophyd.sim import (
    SynGauss,
    Syn2DGauss,
    SynAxis,
    make_fake_device,
    FakeEpicsSignal,
    FakeEpicsSignalRO,
    FakeEpicsSignalWithRBV,
    FakeEpicsPathSignal,
    clear_fake_device,
    instantiate_fake_device,
    SynSignalWithRegistry,
)
from ophyd.device import (
    Device,
    Component as Cpt,
    FormattedComponent as FCpt,
    DynamicDeviceComponent as DDCpt,
)
from ophyd.signal import Signal, EpicsSignal, EpicsSignalRO
from ophyd.areadetector.base import EpicsSignalWithRBV
from ophyd.areadetector.paths import EpicsPathSignal
from ophyd.utils import ReadOnlyError, LimitError, DisconnectedError
import numpy as np
import pytest
import tempfile
import shutil


def test_random_state_gauss1d():
    """With given random state, the output value should stay the same.
    Test performs on 1D gaussian.
    """
    dlist = []
    motor = SynAxis(name="motor")
    for i in range(2):
        s = np.random.RandomState(0)
        noisy_det = SynGauss(
            "noisy_det",
            motor,
            "motor",
            center=0,
            Imax=1,
            noise="uniform",
            sigma=1,
            noise_multiplier=0.1,
            random_state=s,
        )
        noisy_det.trigger()
        d = noisy_det.read()["noisy_det"]["value"]
        dlist.append(d)
    assert dlist[0] == dlist[1]

    # Without random state, output will be different.
    dlist.clear()
    for i in range(2):
        noisy_det = SynGauss(
            "noisy_det",
            motor,
            "motor",
            center=0,
            Imax=1,
            noise="uniform",
            sigma=1,
            noise_multiplier=0.1,
        )
        noisy_det.trigger()
        d = noisy_det.read()["noisy_det"]["value"]
        dlist.append(d)
    assert dlist[0] != dlist[1]


def test_random_state_gauss2d():
    """With given random state, the output value should stay the same.
    Test performs on 2D gaussian.
    """
    dlist = []
    motor1 = SynAxis(name="motor1")
    motor2 = SynAxis(name="motor2")
    for i in range(2):
        s = np.random.RandomState(0)
        noisy_det = Syn2DGauss(
            "noisy_det",
            motor1,
            "motor1",
            motor2,
            "motor2",
            center=(0, 0),
            Imax=1,
            noise="uniform",
            sigma=1,
            noise_multiplier=0.1,
            random_state=s,
        )
        noisy_det.trigger()
        d = noisy_det.read()["noisy_det"]["value"]
        dlist.append(d)
    assert dlist[0] == dlist[1]


def test_synaxis_subcribe():
    hits = dict.fromkeys(["r", "s", "a"], 0)
    vals = dict.fromkeys(["r", "s", "a"], None)

    def p1(tar, value):
        hits[tar] += 1
        vals[tar] = value

    motor = SynAxis(name="motor1")
    # prime the cb cache so these run an subscription
    motor.set(0)
    motor.subscribe(lambda *, value, _tar="a", **kwargs: p1(_tar, value))
    motor.readback.subscribe(lambda *, value, _tar="r", **kwargs: p1(_tar, value))
    motor.setpoint.subscribe(lambda *, value, _tar="s", **kwargs: p1(_tar, value))

    assert vals["r"] == motor.readback.get()
    assert vals["a"] == motor.readback.get()
    assert vals["s"] == motor.setpoint.get()

    assert all(v == 1 for v in hits.values())

    motor.set(1)

    assert vals["r"] == motor.readback.get()
    assert vals["a"] == motor.readback.get()
    assert vals["s"] == motor.setpoint.get()

    assert all(v == 2 for v in hits.values())


def test_synaxis_timestamps():
    from ophyd.status import wait
    import time

    def time_getter(m):
        return {k: v["timestamp"] for k, v in m.read().items()}

    def tester(m, orig_time):
        new_time = time_getter(m)
        assert orig_time != new_time
        return new_time

    motor = SynAxis(name="motor1")
    motor.delay = 0.01
    orig_time = time_getter(motor)

    wait(motor.set(3))
    orig_time = tester(motor, orig_time)

    wait(motor.setpoint.set(4))
    orig_time = tester(motor, orig_time)

    motor.setpoint.put(3)
    time.sleep(2 * motor.delay)
    orig_time = tester(motor, orig_time)


# Classes for testing make_fake_device
class SampleNested(Device):
    yolk = Cpt(EpicsSignal, ":YOLK", string=True)
    whites = Cpt(EpicsSignalRO, ":WHITES")


class Sample(Device):
    egg = Cpt(SampleNested, ":EGG")
    butter = Cpt(
        EpicsSignal,
        ":BUTTER",
        timeout=10.0,
        write_timeout=10.0,
        connection_timeout=10.0,
    )
    flour = Cpt(EpicsSignalRO, ":FLOUR")
    baster = FCpt(EpicsSignal, "{self.drawer}:BASTER")
    sink = FCpt(EpicsSignal, "{self.sink_location}:SINK")
    fridge = DDCpt(
        {"milk": (EpicsSignal, ":MILK", {}), "cheese": (EpicsSignalRO, ":CHEESE", {})}
    )
    nothing = Cpt(Signal)

    def __init__(
        self, prefix, *, drawer="UNDER_THE_SINK", sink_location="COUNTER", **kwargs
    ):
        self.drawer = drawer
        self.sink_location = sink_location
        super().__init__(prefix, **kwargs)


def test_make_fake_device():
    assert make_fake_device(EpicsSignal) == FakeEpicsSignal
    assert make_fake_device(EpicsSignalRO) == FakeEpicsSignalRO
    assert make_fake_device(EpicsSignalWithRBV) == FakeEpicsSignalWithRBV
    assert make_fake_device(EpicsPathSignal) == FakeEpicsPathSignal

    FakeSample = make_fake_device(Sample)
    my_fake = FakeSample("KITCHEN", name="kitchen")
    assert isinstance(my_fake, Sample)

    # Skipped
    assert my_fake.nothing.__class__ is Signal

    # Normal
    assert isinstance(my_fake.butter, FakeEpicsSignal)
    assert isinstance(my_fake.flour, FakeEpicsSignalRO)
    assert isinstance(my_fake.sink, FakeEpicsSignal)

    # Nested
    assert isinstance(my_fake.egg.yolk, FakeEpicsSignal)
    assert isinstance(my_fake.egg.whites, FakeEpicsSignalRO)

    # Dynamic
    assert isinstance(my_fake.fridge.milk, FakeEpicsSignal)
    assert isinstance(my_fake.fridge.cheese, FakeEpicsSignalRO)

    my_fake.read()


def test_clear_fake_device():
    FakeSample = make_fake_device(Sample)
    my_fake = FakeSample("KITCHEN", name="kitchen")
    clear_fake_device(my_fake, default_value=49, default_string_value="string")
    assert my_fake.butter.get() == 49
    assert my_fake.flour.get() == 49
    assert my_fake.sink.get() == 49
    assert my_fake.egg.yolk.get() == "string"
    assert my_fake.egg.whites.get() == 49


def test_instantiate_fake_device():
    my_fake = instantiate_fake_device(Sample)
    assert my_fake.drawer == "UNDER_THE_SINK"
    assert my_fake.sink_location == "COUNTER"
    assert my_fake.name == "FakeSample"
    assert my_fake.prefix == "_prefix"

    my_fake = instantiate_fake_device(Sample, drawer="JUNK_DRAWER")
    assert my_fake.drawer == "JUNK_DRAWER"
    assert my_fake.sink_location == "COUNTER"
    assert my_fake.name == "FakeSample"


def test_do_not_break_real_class():
    make_fake_device(Sample)
    assert Sample.butter.cls is EpicsSignal
    assert Sample.egg.cls is SampleNested
    assert SampleNested.whites.cls is EpicsSignalRO
    assert Sample.fridge.defn["milk"][0] is EpicsSignal

    with pytest.raises(DisconnectedError):
        my_real = Sample("KITCHEN", name="kitchen")
        my_real.read()


def test_fake_epics_signal():
    sig = FakeEpicsSignal("PVNAME", name="sig", limits=True)
    with pytest.raises(ValueError):
        sig.put(None)
    sig.sim_set_limits((0, 10))
    with pytest.raises(LimitError):
        sig.put(11)
    sig.put(4)
    assert sig.get() == 4
    sig.sim_put(5)
    assert sig.get() == 5
    sig.sim_set_putter(lambda x: sig.sim_put(x + 1))
    sig.put(6)
    assert sig.get() == 7
    assert sig.get(as_string=True) == str(7)


def test_fake_epics_signal_ro():
    sig = FakeEpicsSignalRO("PVNAME", name="sig")
    with pytest.raises(ReadOnlyError):
        sig.put(3)
    with pytest.raises(ReadOnlyError):
        sig.put(4)
    with pytest.raises(ReadOnlyError):
        sig.set(5)
    sig.sim_put(1)
    assert sig.get() == 1


def test_fake_epics_signal_enum():
    sig = FakeEpicsSignal("PVNAME", name="sig", string=True)
    sig.sim_set_enum_strs(["zero", "one", "two", "three"])
    sig.put(0)
    assert sig.describe()["sig"]["enum_strs"] == ("zero", "one", "two", "three")
    assert sig.get() == "zero"
    assert sig.get(as_string=False) == 0
    sig.put("two")
    assert sig.get(as_string=False) == 2
    with pytest.raises(ValueError):
        sig.put("bazillion")


def test_SynSignalWithRegistry():
    tempdirname = tempfile.mkdtemp()

    def data_func():
        return np.array(np.ones((10, 10)))

    img = SynSignalWithRegistry(
        data_func, save_path=tempdirname, name="img", labels={"detectors"}
    )
    img.stage()
    img.trigger()
    d0 = img.read()
    assert int(d0["img"]["value"][-1]) == 0
    img.trigger()
    d1 = img.read()
    assert int(d1["img"]["value"][-1]) == 1  # increased by 1
    shutil.rmtree(tempdirname)


def test_synaxis_describe():
    bs = pytest.importorskip("bluesky")
    import bluesky.plans as bp

    motor1 = SynAxis(name="motor1")
    RE = bs.RunEngine()
    RE(bp.scan([], motor1, -5, 5, 5))


def test_describe(hw):
    # These need to be staged and triggered before they can be described, just
    # like real area detectors do. We plan to change this approach and remove
    # this limitation in ophyd 1.6.0, but for now we'll just skip these.
    SKIP = (
        "img",
        "direct_img",
        "direct_img_list",
    )
    for name, obj in hw.__dict__.items():
        if name in SKIP:
            continue
        if hasattr(obj, "describe"):
            obj.describe()
        elif hasattr(obj, "describe_collect"):
            obj.describe_collect()
        else:
            raise AttributeError("expected describe or describe_collect")



#!/usr/bin/env python3
from caproto.server import pvproperty, PVGroup, ioc_arg_parser, run
import numpy as np


class FakeMotorIOC(PVGroup):
    setpoint = pvproperty(value=0.0, precision=1)
    readback = pvproperty(value=0.0, read_only=True, precision=1)
    moving = pvproperty(value=0.0, read_only=True)
    actuate = pvproperty(value=0)
    stop = pvproperty(value=0)
    step_size = pvproperty(value=0.1)

    @actuate.scan(period=0.1)
    async def actuate(self, instance, async_lib):
        step_size = self.step_size.value
        setpoint = self.setpoint.value
        readback = self.readback.value
        moving = self.moving.value
        actuate = self.actuate.value
        stop = self.stop.value

        if stop:
            await self.stop.write(0)
            await self.moving.write(0)
        elif actuate or moving:
            if moving != 1:
                await self.actuate.write(0)
                await self.moving.write(1)

            delta = setpoint - readback
            if abs(delta) <= step_size:
                await self.readback.write(setpoint)
                await self.moving.write(0)
            else:
                await self.readback.write(readback + np.sign(delta) * step_size)


if __name__ == "__main__":
    ioc_options, run_options = ioc_arg_parser(
        default_prefix="fake_motor:", desc="An IOC which mocks a simple motor"
    )
    ioc = FakeMotorIOC(**ioc_options)
    run(ioc.pvdb, **run_options)
